# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.api.v2 import attributes as attrs
from neutron.common import exceptions as n_exc
from neutron.db import portbindings_db
from neutron.extensions import portbindings

from networking_nec.plugins.openflow.db import api as ndb
from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow import utils as necutils


class PortBindingMixin(portbindings_db.PortBindingMixin):

    def _get_base_binding_dict(self):
        sg_enabled = sg_rpc.is_firewall_enabled()
        vif_details = {portbindings.CAP_PORT_FILTER: sg_enabled,
                       portbindings.OVS_HYBRID_PLUG: sg_enabled}
        binding = {portbindings.VIF_TYPE: portbindings.VIF_TYPE_OVS,
                   portbindings.VIF_DETAILS: vif_details}
        return binding

    def _extend_port_dict_binding_portinfo(self, port_res, portinfo):
        if portinfo:
            port_res[portbindings.PROFILE] = {
                'datapath_id': portinfo['datapath_id'],
                'port_no': portinfo['port_no'],
            }
        elif portbindings.PROFILE in port_res:
            del port_res[portbindings.PROFILE]

    def _validate_portinfo(self, profile):
        key_specs = {
            'datapath_id': {'type:string': None, 'required': True},
            'port_no': {'type:non_negative': None, 'required': True,
                        'convert_to': attrs.convert_to_int}
        }
        msg = attrs._validate_dict_or_empty(profile, key_specs=key_specs)
        if msg:
            raise n_exc.InvalidInput(error_message=msg)

        datapath_id = profile.get('datapath_id')
        port_no = profile.get('port_no')
        try:
            dpid = int(datapath_id, 16)
        except ValueError:
            raise nexc.ProfilePortInfoInvalidDataPathId()
        if dpid > 0xffffffffffffffffL:
            raise nexc.ProfilePortInfoInvalidDataPathId()
        # Make sure dpid is a hex string beginning with 0x.
        dpid = hex(dpid)

        if int(port_no) > 65535:
            raise nexc.ProfilePortInfoInvalidPortNo()

        return {'datapath_id': dpid, 'port_no': port_no}

    def _process_portbindings_portinfo_create(self, context, port_data, port):
        """Add portinfo according to bindings:profile in create_port().

        :param context: neutron api request context
        :param port_data: port attributes passed in PUT request
        :param port: port attributes to be returned
        """
        profile = port_data.get(portbindings.PROFILE)
        # If portbindings.PROFILE is None, unspecified or an empty dict
        # it is regarded that portbinding.PROFILE is not set.
        profile_set = attrs.is_attr_set(profile) and profile
        if profile_set:
            portinfo = self._validate_portinfo(profile)
            portinfo['mac'] = port['mac_address']
            ndb.add_portinfo(context.session, port['id'], **portinfo)
        else:
            portinfo = None
        self._extend_port_dict_binding_portinfo(port, portinfo)

    def _process_portbindings_portinfo_update(self, context, port_data, port):
        """Update portinfo according to bindings:profile in update_port().

        :param context: neutron api request context
        :param port_data: port attributes passed in PUT request
        :param port: port attributes to be returned
        """
        if portbindings.PROFILE not in port_data:
            return
        profile = port_data.get(portbindings.PROFILE)
        # If binding:profile is None or an empty dict,
        # it means binding:.profile needs to be cleared.
        # TODO(amotoki): Allow Make None in binding:profile in
        # the API layer. See LP bug #1220011.
        profile_set = attrs.is_attr_set(profile) and profile
        cur_portinfo = ndb.get_portinfo(context.session, port['id'])
        if profile_set:
            portinfo = self._validate_portinfo(profile)
            if cur_portinfo:
                if (necutils.cmp_dpid(portinfo['datapath_id'],
                                      cur_portinfo.datapath_id) and
                        portinfo['port_no'] == cur_portinfo.port_no):
                    return
                ndb.del_portinfo(context.session, port['id'])
            portinfo['mac'] = port['mac_address']
            ndb.add_portinfo(context.session, port['id'], **portinfo)
        elif cur_portinfo:
            portinfo = None
            ndb.del_portinfo(context.session, port['id'])
        else:
            portinfo = None
        self._extend_port_dict_binding_portinfo(port, portinfo)

    def extend_port_dict_binding(self, port_res, port_db):
        super(PortBindingMixin, self).extend_port_dict_binding(port_res,
                                                               port_db)
        self._extend_port_dict_binding_portinfo(port_res, port_db.portinfo)

    def _process_portbindings_create(self, context, port_data, port):
        super(PortBindingMixin, self)._process_portbindings_create_and_update(
            context, port_data, port)
        self._process_portbindings_portinfo_create(context, port_data, port)

    def _process_portbindings_update(self, context, port_data, port):
        super(PortBindingMixin, self)._process_portbindings_create_and_update(
            context, port_data, port)
        self._process_portbindings_portinfo_update(context, port_data, port)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import string
import uuid

import mock
import netaddr
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.common import constants
from neutron.tests import base

from networking_nec.plugins.openflow.db import models as nmodels
from networking_nec.plugins.openflow import drivers
from networking_nec.plugins.openflow.drivers import pfc
from networking_nec.plugins.openflow.extensions import packetfilter as ext_pf
from networking_nec.plugins.openflow import ofc_client as ofc


class TestConfig(object):
    """Configuration for this test."""
    host = '127.0.0.1'
    port = 8888
    use_ssl = False
    key_file = None
    cert_file = None
    insecure_ssl = False


def _ofc(id):
    """OFC ID converter."""
    return "ofc-%s" % id


class PFCDriverTestBase(base.BaseTestCase):

    driver = 'networking_nec.plugins.openflow.drivers.pfc.PFCDriverBase'
    filter_supported = False

    def setUp(self):
        super(PFCDriverTestBase, self).setUp()
        self.driver = drivers.get_driver(self.driver)(TestConfig)
        self.do_request = mock.patch.object(ofc.OFCClient,
                                            'do_request').start()

    def get_ofc_item_random_params(self):
        """create random parameters for ofc_item test."""
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        portinfo = nmodels.PortInfo(id=port_id, datapath_id="0x123456789",
                                    port_no=1234, vlan_id=321,
                                    mac="11:22:33:44:55:66")
        return tenant_id, network_id, portinfo

    def _generate_ofc_tenant_id(self, tenant_id):
        fields = tenant_id.split('-')
        # Strip 1st character (UUID version) of 3rd field
        fields[2] = fields[2][1:]
        return ''.join(fields)

    def get_ofc_description(self, desc):
        """OFC description consists of [A-Za-z0-9_]."""
        return desc.replace('-', '_').replace(' ', '_')

    def _create_tenant(self, t, ofc_t, post_id=False, post_desc=False):
        tenant_path = '/tenants/%s' % ofc_t
        path = "/tenants"
        description = "desc of %s" % t
        body = {}
        if post_desc:
            ofc_description = self.get_ofc_description(description)
            body['description'] = ofc_description
        if post_id:
            body['id'] = ofc_t
            self.do_request.return_value = None
        else:
            self.do_request.return_value = {'id': ofc_t}

        ret = self.driver.create_tenant(description, t)
        self.do_request.assert_called_once_with("POST", path, body=body)
        self.assertEqual(ret, tenant_path)

    def testa_create_tenant(self):
        t, n, p = self.get_ofc_item_random_params()
        ofc_t = self._generate_ofc_tenant_id(t)
        self._create_tenant(t, ofc_t, post_id=True)

    def testc_delete_tenant(self):
        t, n, p = self.get_ofc_item_random_params()

        path = "/tenants/%s" % _ofc(t)

        self.driver.delete_tenant(path)
        self.do_request.assert_called_once_with("DELETE", path)

    def testd_create_network(self):
        t, n, p = self.get_ofc_item_random_params()
        description = "desc of %s" % n
        ofc_description = self.get_ofc_description(description)

        tenant_path = "/tenants/%s" % _ofc(t)
        post_path = "%s/networks" % tenant_path
        body = {'description': ofc_description}
        network = {'id': _ofc(n)}
        self.do_request.return_value = network

        ret = self.driver.create_network(tenant_path, description, n)
        self.do_request.assert_called_once_with("POST", post_path, body=body)
        net_path = "/tenants/%s/networks/%s" % (_ofc(t), _ofc(n))
        self.assertEqual(ret, net_path)

    def testf_delete_network(self):
        t, n, p = self.get_ofc_item_random_params()

        net_path = "/tenants/%s/networks/%s" % (_ofc(t), _ofc(n))

        self.driver.delete_network(net_path)
        self.do_request.assert_called_once_with("DELETE", net_path)

    def _test_create_port(self, call_filters_arg=None, send_filters_arg=None):
        t, n, p = self.get_ofc_item_random_params()

        net_path = "/tenants/%s/networks/%s" % (_ofc(t), _ofc(n))
        post_path = "%s/ports" % net_path
        port_path = "/tenants/%s/networks/%s/ports/%s" % (_ofc(t), _ofc(n),
                                                          _ofc(p.id))
        body = {'datapath_id': p.datapath_id,
                'port': str(p.port_no),
                'vid': str(p.vlan_id)}
        if send_filters_arg is not None:
            body['filters'] = send_filters_arg
        port = {'id': _ofc(p.id)}
        self.do_request.return_value = port

        if call_filters_arg is not None:
            ret = self.driver.create_port(net_path, p, p.id, call_filters_arg)
        else:
            ret = self.driver.create_port(net_path, p, p.id)
        self.do_request.assert_called_once_with("POST", post_path, body=body)
        self.assertEqual(ret, port_path)

    def testg_create_port(self):
        self._test_create_port()

    def test_create_port_with_filters_argument(self):
        # If no filter support, 'filters' argument is passed to OFC.
        # Note that it will be overridden in a test class with filter support.
        self._test_create_port(call_filters_arg=['dummy'],
                               send_filters_arg=None)

    def testh_delete_port(self):
        t, n, p = self.get_ofc_item_random_params()

        port_path = "/tenants/%s/networks/%s/ports/%s" % (_ofc(t), _ofc(n),
                                                          _ofc(p.id))

        self.driver.delete_port(port_path)
        self.do_request.assert_called_once_with("DELETE", port_path)

    def test_filter_supported(self):
        self.assertEqual(self.filter_supported, self.driver.filter_supported())


class PFCDriverBaseTest(PFCDriverTestBase):

    def test_extract_ofc_network_id(self):
        network_id = '/tenants/tenant-a/networks/network-a'
        self.assertEqual('network-a',
                         self.driver._extract_ofc_network_id(network_id))

    def test_extract_ofc_network_id_failure(self):
        network_id = '/tenants/tenant-a/networks/network-a/dummy'
        self.assertRaises(pfc.InvalidOFCIdFormat,
                          self.driver._extract_ofc_network_id, network_id)

    def test_extract_ofc_port_id(self):
        port_id = '/tenants/tenant-a/networks/network-a/ports/port-a'
        self.assertEqual({'tenant': 'tenant-a',
                          'network': 'network-a',
                          'port': 'port-a'},
                         self.driver._extract_ofc_port_id(port_id))

    def test_extract_ofc_port_id_failure(self):
        port_id = '/tenants/tenant-a/dummy/network-a/ports/port-a'
        self.assertRaises(pfc.InvalidOFCIdFormat,
                          self.driver._extract_ofc_port_id, port_id)


class PFCV3DriverTest(PFCDriverTestBase):
    driver = 'pfc_v3'

    def testa_create_tenant(self):
        t, n, p = self.get_ofc_item_random_params()
        ret = self.driver.create_tenant('dummy_desc', t)
        self.assertEqual(0, self.do_request.call_count)
        ofc_t_path = "/tenants/" + self._generate_ofc_tenant_id(t)
        self.assertEqual(ofc_t_path, ret)

    def testc_delete_tenant(self):
        t, n, p = self.get_ofc_item_random_params()
        path = "/tenants/%s" % _ofc(t)
        self.driver.delete_tenant(path)
        self.assertEqual(0, self.do_request.call_count)


class PFCV4DriverTest(PFCDriverTestBase):
    driver = 'pfc_v4'


class PFCV5DriverTest(PFCDriverTestBase):
    driver = 'pfc_v5'

    def test_create_router(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()
        description = 'dummy_router_desc'

        tenant_path = "/tenants/%s" % _ofc(t)
        post_path = "%s/routers" % tenant_path
        router = {'id': _ofc(r)}
        self.do_request.return_value = router

        ret = self.driver.create_router(tenant_path, description, r)
        self.do_request.assert_called_once_with("POST", post_path, body=None)
        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        self.assertEqual(ret, router_path)

    def test_delete_router(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))

        self.driver.delete_router(router_path)
        self.do_request.assert_called_once_with("DELETE", router_path)

    def test_add_router_interface(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()
        n = uuidutils.generate_uuid()
        p = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        infs_path = router_path + "/interfaces"
        net_path = "/tenants/%s/networks/%s" % (_ofc(t), _ofc(n))
        ip_address = '10.1.1.1/24'
        mac_address = '11:22:33:44:55:66'
        body = {'net_id': _ofc(n),
                'ip_address': ip_address,
                'mac_address': mac_address}
        inf = {'id': _ofc(p)}
        self.do_request.return_value = inf

        ret = self.driver.add_router_interface(router_path, net_path,
                                               ip_address, mac_address)
        self.do_request.assert_called_once_with("POST", infs_path, body=body)

        inf_path = "%s/interfaces/%s" % (router_path, _ofc(p))
        self.assertEqual(ret, inf_path)

    def test_update_router_interface(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()
        p = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        inf_path = "%s/interfaces/%s" % (router_path, _ofc(p))
        ip_address = '10.1.1.1/24'
        mac_address = '11:22:33:44:55:66'

        self.driver.update_router_interface(inf_path, ip_address, mac_address)
        self.driver.update_router_interface(inf_path, ip_address=ip_address)
        self.driver.update_router_interface(inf_path, mac_address=mac_address)

        self.do_request.assert_has_calls([
            mock.call("PUT", inf_path, body={'ip_address': ip_address,
                                             'mac_address': mac_address}),
            mock.call("PUT", inf_path, body={'ip_address': ip_address}),
            mock.call("PUT", inf_path, body={'mac_address': mac_address}),
        ])

    def test_delete_router_interface(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()
        p = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        inf_path = "%s/interfaces/%s" % (router_path, _ofc(p))

        self.driver.delete_router_interface(inf_path)
        self.do_request.assert_called_once_with("DELETE", inf_path)

    def _get_route_id(self, dest, nexthop):
        dest = netaddr.IPNetwork(dest)
        return '-'.join([str(dest.network), nexthop, str(dest.netmask)])

    def test_add_router_route(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        routes_path = router_path + "/routes"
        dest = '10.1.1.0/24'
        nexthop = '192.168.100.10'
        body = {'destination': dest, 'nexthop': nexthop}
        route_id = self._get_route_id(dest, nexthop)
        self.do_request.return_value = {'id': route_id}

        ret = self.driver.add_router_route(router_path, '10.1.1.0/24',
                                           '192.168.100.10')
        self.do_request.assert_called_once_with("POST", routes_path, body=body)
        route_path = routes_path + '/' + route_id
        self.assertEqual(ret, route_path)

    def test_delete_router_route(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        routes_path = router_path + "/routes"

        route_id = self._get_route_id('10.1.1.0/24', '192.168.100.10')
        route_path = routes_path + '/' + route_id

        self.driver.delete_router_route(route_path)
        self.do_request.assert_called_once_with("DELETE", route_path)

    def test_list_router_routes(self):
        t = uuidutils.generate_uuid()
        r = uuidutils.generate_uuid()

        router_path = "/tenants/%s/routers/%s" % (_ofc(t), _ofc(r))
        routes_path = router_path + "/routes"

        routes = [('10.1.1.0/24', '192.168.100.10'),
                  ('10.2.2.0/20', '192.168.100.20')]
        data = {'routes': [{'id': self._get_route_id(route[0], route[1]),
                            'destination': route[0], 'nexthop': route[1]}
                           for route in routes]}
        self.do_request.return_value = data

        ret = self.driver.list_router_routes(router_path)
        self.do_request.assert_called_once_with("GET", routes_path)

        expected = [{'id': (routes_path + "/" +
                            self._get_route_id(route[0], route[1])),
                     'destination': route[0], 'nexthop': route[1]}
                    for route in routes]
        self.assertEqual(len(routes), len(ret))
        self.assertEqual(data['routes'], expected)


class PFCFilterDriverTestMixin(object):
    def _test_create_filter(self, filter_dict=None, filter_post=None,
                            apply_ports=None):
        t, n, p = self.get_ofc_item_random_params()

        filter_id = uuidutils.generate_uuid()
        f = {'priority': 123, 'action': "ACCEPT",
             'network_id': n}
        if filter_dict:
            f.update(filter_dict)

        body = {'action': 'pass', 'priority': 123}
        if filter_post:
            body.update(filter_post)

        self.do_request.return_value = {'id': filter_id}
        with mock.patch('networking_nec.plugins.openflow.db.api.'
                        'get_active_ports_on_ofc',
                        return_value=apply_ports) as active_ports:
            ret = self.driver.create_filter(mock.sentinel.ctx, f)
        self.do_request.assert_called_once_with("POST", "/filters",
                                                body=body)
        self.assertEqual(ret, '/filters/%s' % filter_id)
        active_ports.assert_called_once_with(mock.sentinel.ctx, n,
                                             filter_dict.get('in_port'))

    def test_create_filter_accept(self):
        self._test_create_filter(filter_dict={'action': 'ACCEPT'})

    def test_create_filter_allow(self):
        self._test_create_filter(filter_dict={'action': 'ALLOW'})

    def test_create_filter_deny(self):
        self._test_create_filter(filter_dict={'action': 'DENY'},
                                 filter_post={'action': 'drop'})

    def test_create_filter_drop(self):
        self._test_create_filter(filter_dict={'action': 'DROP'},
                                 filter_post={'action': 'drop'})

    def test_create_filter_empty_field_not_post(self):
        filter_dict = {'src_mac': '', 'src_cidr': '', 'src_port': 0,
                       'dst_mac': '', 'dst_cidr': '', 'dst_port': 0,
                       'protocol': '', 'eth_type': 0}
        filter_post = {}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_none_field_not_post(self):
        filter_dict = {'src_mac': None, 'src_cidr': None, 'src_port': None,
                       'dst_mac': None, 'dst_cidr': None, 'dst_port': None,
                       'protocol': None, 'eth_type': None}
        filter_post = {}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_all_fields(self):
        filter_dict = {'src_mac': '11:22:33:44:55:66',
                       'dst_mac': '77:88:99:aa:bb:cc',
                       'src_cidr': '192.168.3.0/24',
                       'dst_cidr': '10.11.240.0/20',
                       'src_port': 12345,
                       'dst_port': 23456,
                       'protocol': '0x10',
                       'eth_type': 0x800}
        filter_post = filter_dict.copy()
        filter_post['protocol'] = 16
        filter_post['eth_type'] = '0x800'
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_cidr_ip_addr_32(self):
        filter_dict = {'src_cidr': '192.168.3.1',
                       'dst_cidr': '10.11.240.2'}
        filter_post = {'src_cidr': '192.168.3.1/32',
                       'dst_cidr': '10.11.240.2/32'}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_proto_tcp(self):
        filter_dict = {'protocol': 'TCP'}
        filter_post = {'protocol': constants.PROTO_NUM_TCP}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_proto_udp(self):
        filter_dict = {'protocol': 'UDP'}
        filter_post = {'protocol': constants.PROTO_NUM_UDP}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_proto_icmp(self):
        filter_dict = {'protocol': 'ICMP'}
        filter_post = {'protocol': constants.PROTO_NUM_ICMP}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_proto_arp_not_proto_post(self):
        filter_dict = {'protocol': 'ARP'}
        filter_post = {}
        self._test_create_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_create_filter_apply_ports(self):
        apply_ports = [
            ('p1', '/tenants/tenant-1/networks/network-1/ports/port-1'),
            ('p2', '/tenants/tenant-2/networks/network-2/ports/port-2')]
        filter_post = {'apply_ports': [
            {'tenant': 'tenant-1', 'network': 'network-1', 'port': 'port-1'},
            {'tenant': 'tenant-2', 'network': 'network-2', 'port': 'port-2'}
        ]}
        self._test_create_filter(filter_dict={}, apply_ports=apply_ports,
                                 filter_post=filter_post)

    def test_create_filter_apply_ports_with_router_interface(self):
        apply_ports = [
            ('p1', '/tenants/tenant-1/networks/network-1/ports/port-1'),
            ('p2', '/tenants/tenant-2/routers/router-3/interfaces/if-4')]
        filter_post = {'apply_ports': [
            {'tenant': 'tenant-1', 'network': 'network-1', 'port': 'port-1'},
            {'tenant': 'tenant-2', 'router': 'router-3', 'interface': 'if-4'}
        ]}
        self._test_create_filter(filter_dict={}, apply_ports=apply_ports,
                                 filter_post=filter_post)

    def test_create_filter_apply_ports_no_router_interface_support(self):
        cfg.CONF.set_override('support_packet_filter_on_ofc_router',
                              False, 'OFC')
        apply_ports = [
            ('p1', '/tenants/tenant-1/networks/network-1/ports/port-1'),
            ('p2', '/tenants/tenant-2/routers/router-3/interfaces/if-4')]
        filter_post = {'apply_ports': [
            {'tenant': 'tenant-1', 'network': 'network-1', 'port': 'port-1'},
        ]}
        self._test_create_filter(filter_dict={}, apply_ports=apply_ports,
                                 filter_post=filter_post)

    def _test_update_filter(self, filter_dict=None, filter_post=None):
        filter_id = uuidutils.generate_uuid()
        ofc_filter_id = '/filters/%s' % filter_id
        self.driver.update_filter(ofc_filter_id, filter_dict)
        self.do_request.assert_called_once_with("PUT", ofc_filter_id,
                                                body=filter_post)

    def test_update_filter_empty_fields(self):
        filter_dict = {'src_mac': '', 'src_cidr': '', 'src_port': 0,
                       'dst_mac': '', 'dst_cidr': '', 'dst_port': 0,
                       'protocol': '', 'eth_type': 0}
        filter_post = {'src_mac': '', 'src_cidr': '', 'src_port': '',
                       'dst_mac': '', 'dst_cidr': '', 'dst_port': '',
                       'protocol': '', 'eth_type': ''}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_none_fields(self):
        filter_dict = {'src_mac': None, 'src_cidr': None, 'src_port': None,
                       'dst_mac': None, 'dst_cidr': None, 'dst_port': None,
                       'protocol': None, 'eth_type': None}
        filter_post = {'src_mac': '', 'src_cidr': '', 'src_port': '',
                       'dst_mac': '', 'dst_cidr': '', 'dst_port': '',
                       'protocol': '', 'eth_type': ''}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_all_fields(self):
        filter_dict = {'src_mac': '11:22:33:44:55:66',
                       'dst_mac': '77:88:99:aa:bb:cc',
                       'src_cidr': '192.168.3.0/24',
                       'dst_cidr': '10.11.240.0/20',
                       'src_port': 12345,
                       'dst_port': 23456,
                       'protocol': '0x10',
                       'eth_type': 0x800}
        filter_post = filter_dict.copy()
        filter_post['protocol'] = 16
        filter_post['eth_type'] = '0x800'
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_cidr_ip_addr_32(self):
        filter_dict = {'src_cidr': '192.168.3.1',
                       'dst_cidr': '10.11.240.2'}
        filter_post = {'src_cidr': '192.168.3.1/32',
                       'dst_cidr': '10.11.240.2/32'}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_proto_tcp(self):
        filter_dict = {'protocol': 'TCP'}
        filter_post = {'protocol': constants.PROTO_NUM_TCP}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_proto_udp(self):
        filter_dict = {'protocol': 'UDP'}
        filter_post = {'protocol': constants.PROTO_NUM_UDP}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_proto_icmp(self):
        filter_dict = {'protocol': 'ICMP'}
        filter_post = {'protocol': constants.PROTO_NUM_ICMP}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_update_filter_proto_arp_post_empty(self):
        filter_dict = {'protocol': 'ARP'}
        filter_post = {'protocol': ''}
        self._test_update_filter(filter_dict=filter_dict,
                                 filter_post=filter_post)

    def test_delete_filter(self):
        t, n, p = self.get_ofc_item_random_params()
        f_path = "/filters/%s" % uuidutils.generate_uuid()
        self.driver.delete_filter(f_path)
        self.do_request.assert_called_once_with("DELETE", f_path)

    def _test_validate_filter_duplicate_priority(self, method, found_dup):
        with mock.patch('neutron.manager.NeutronManager'
                        '.get_plugin') as get_plugin:
            plugin = get_plugin.return_value
            if found_dup:
                plugin.get_packet_filters.return_value = ['found']
            else:
                plugin.get_packet_filters.return_value = []
            network_id = str(uuid.uuid4())
            filter_dict = {'network_id': network_id,
                           'priority': 12}
            if found_dup:
                self.assertRaises(ext_pf.PacketFilterDuplicatedPriority,
                                  method, 'context', filter_dict)
            else:
                self.assertIsNone(method('context', filter_dict))
            plugin.get_packet_filters.assert_called_once_with(
                'context',
                filters={'network_id': [network_id],
                         'priority': [12]},
                fields=['id'])

    def test_validate_filter_create_no_duplicate_priority(self):
        self._test_validate_filter_duplicate_priority(
            self.driver.validate_filter_create,
            found_dup=False)

    def test_validate_filter_create_duplicate_priority(self):
        self._test_validate_filter_duplicate_priority(
            self.driver.validate_filter_create,
            found_dup=True)

    def test_validate_filter_update_action_raises_error(self):
        filter_dict = {'action': 'ALLOW'}
        self.assertRaises(ext_pf.PacketFilterUpdateNotSupported,
                          self.driver.validate_filter_update,
                          'context', filter_dict)

    def test_validate_filter_update_priority_raises_error(self):
        filter_dict = {'priority': '13'}
        self.assertRaises(ext_pf.PacketFilterUpdateNotSupported,
                          self.driver.validate_filter_update,
                          'context', filter_dict)

    def _test_validate_filter_ipv6_not_supported(self, field, create=True):
        if create:
            filter_dict = {'network_id': 'net1', 'priority': 12}
            method = self.driver.validate_filter_create
        else:
            filter_dict = {}
            method = self.driver.validate_filter_update
        filter_dict[field] = 'fe80::1'
        self.assertRaises(ext_pf.PacketFilterIpVersionNonSupported,
                          method, 'context', filter_dict)
        filter_dict[field] = '10.56.3.3'
        self.assertIsNone(method('context', filter_dict))

    def test_validate_filter_create_ipv6_not_supported(self):
        with mock.patch('neutron.manager.NeutronManager'
                        '.get_plugin') as get_plugin:
            plugin = get_plugin.return_value
            plugin.get_packet_filters.return_value = []
            self._test_validate_filter_ipv6_not_supported(
                'src_cidr', create=True)
            self._test_validate_filter_ipv6_not_supported(
                'dst_cidr', create=True)

    def test_validate_filter_update_ipv6_not_supported(self):
        self._test_validate_filter_ipv6_not_supported('src_cidr', create=False)
        self._test_validate_filter_ipv6_not_supported('dst_cidr', create=False)

    def _test_validate_filter_priority_range_one(self, method, priority, ok):
        filter_dict = {'priority': priority, 'network_id': 'net1'}
        if ok:
            self.assertIsNone(method('context', filter_dict))
        else:
            self.assertRaises(ext_pf.PacketFilterInvalidPriority,
                              method, 'context', filter_dict)

    def test_validate_filter_create_priority_range(self):
        with mock.patch('neutron.manager.NeutronManager'
                        '.get_plugin') as get_plugin:
            plugin = get_plugin.return_value
            plugin.get_packet_filters.return_value = []

            method = self.driver.validate_filter_create
            self._test_validate_filter_priority_range_one(method, 0, False)
            self._test_validate_filter_priority_range_one(method, 1, True)
            self._test_validate_filter_priority_range_one(method, 32766, True)
            self._test_validate_filter_priority_range_one(method, 32767, False)


class PFCV51DriverTest(PFCFilterDriverTestMixin, PFCV5DriverTest):
    driver = 'pfc_v51'
    filter_supported = True

    def test_create_port_with_filters_argument(self):
        self._test_create_port(
            call_filters_arg=[('neutron-id-1', '/filters/filter-1'),
                              ('neutron-id-2', '/filters/filter-2')],
            send_filters_arg=['filter-1', 'filter-2'])


class PFCDriverStringTest(base.BaseTestCase):

    driver = 'networking_nec.plugins.openflow.drivers.pfc.PFCDriverBase'

    def setUp(self):
        super(PFCDriverStringTest, self).setUp()
        self.driver = drivers.get_driver(self.driver)(TestConfig)

    def test_generate_pfc_id_uuid(self):
        id_str = uuidutils.generate_uuid()
        exp_str = (id_str[:14] + id_str[15:]).replace('-', '')[:31]

        ret_str = self.driver._generate_pfc_id(id_str)
        self.assertEqual(exp_str, ret_str)

    def test_generate_pfc_id_uuid_no_hyphen(self):
        # Keystone tenant_id style uuid
        id_str = uuidutils.generate_uuid()
        id_no_hyphen = id_str.replace('-', '')
        exp_str = (id_str[:14] + id_str[15:]).replace('-', '')[:31]

        ret_str = self.driver._generate_pfc_id(id_no_hyphen)
        self.assertEqual(exp_str, ret_str)

    def test_generate_pfc_id_string(self):
        id_str = uuidutils.generate_uuid() + 'x'
        exp_str = id_str[:31].replace('-', '_')

        ret_str = self.driver._generate_pfc_id(id_str)
        self.assertEqual(exp_str, ret_str)

    def test_generate_pfc_desc(self):
        random_list = [random.choice(string.printable) for x in range(128)]
        random_str = ''.join(random_list)

        accept_letters = string.letters + string.digits
        exp_list = [x if x in accept_letters else '_' for x in random_list]
        exp_str = ''.join(exp_list)[:127]

        ret_str = self.driver._generate_pfc_description(random_str)
        self.assertEqual(exp_str, ret_str)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class OFCDriverBase(object):
    """OpenFlow Controller (OFC) Driver Specification.

    OFCDriverBase defines the minimum set of methods required by this plugin.
    It would be better that other methods like update_* are implemented.
    """

    @abc.abstractmethod
    def create_tenant(self, description, tenant_id=None):
        """Create a new tenant at OpenFlow Controller.

        :param description: A description of this tenant.
        :param tenant_id: A hint of OFC tenant ID.
                          A driver could use this id as a OFC id or ignore it.
        :returns: ID of the tenant created at OpenFlow Controller.
        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass

    @abc.abstractmethod
    def delete_tenant(self, ofc_tenant_id):
        """Delete a tenant at OpenFlow Controller.

        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass

    @abc.abstractmethod
    def create_network(self, ofc_tenant_id, description, network_id=None):
        """Create a new network on specified OFC tenant at OpenFlow Controller.

        :param ofc_tenant_id: a OFC tenant ID in which a new network belongs.
        :param description: A description of this network.
        :param network_id: A hint of an ID of OFC network.
        :returns: ID of the network created at OpenFlow Controller.
            ID returned must be unique in the OpenFlow Controller.
            If a network is identified in conjunction with other information
            such as a tenant ID, such information should be included in the ID.
        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass

    @abc.abstractmethod
    def delete_network(self, ofc_network_id):
        """Delete a netwrok at OpenFlow Controller.

        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass

    @abc.abstractmethod
    def create_port(self, ofc_network_id, portinfo,
                    port_id=None, filters=None):
        """Create a new port on specified network at OFC.

        :param ofc_network_id: a OFC tenant ID in which a new port belongs.
        :param portinfo: An OpenFlow information of this port.
                    {'datapath_id': Switch ID that a port connected.
                     'port_no': Port Number that a port connected on a Swtich.
                     'vlan_id': VLAN ID that a port tagging.
                     'mac': Mac address.
                    }
        :param port_id: A hint of an ID of OFC port.
            ID returned must be unique in the OpenFlow Controller.

            If a port is identified in combination with a network or
            a tenant, such information should be included in the ID.
        :param filters: A list of packet filter associated with the port.
            Each element is a tuple (neutron ID, OFC ID)

        :returns: ID of the port created at OpenFlow Controller.
        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass

    @abc.abstractmethod
    def delete_port(self, ofc_port_id):
        """Delete a port at OpenFlow Controller.

        :raises: neutron.plugin.nec.common.exceptions.OFCException
        """
        pass



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg
import webob.exc

from neutron.common import constants
from neutron.common import topics
from neutron import context
from neutron.db import db_base_plugin_v2
from neutron import manager
from neutron.tests.unit.db import test_allowedaddresspairs_db as test_pair
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin

from networking_nec.plugins.openflow import config as nec_config
from networking_nec.plugins.openflow.db import api as ndb
from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow import rpc
from networking_nec.tests import base
from networking_nec.tests.unit.openflow import fake_ofc_manager


PLUGIN_NAME = 'networking_nec.plugins.openflow.plugin.NECPluginV2'
PLUGIN_BASE = 'networking_nec.plugins.openflow.'
OFC_MANAGER = PLUGIN_BASE + 'ofc_manager.OFCManager'
NOTIFIER = PLUGIN_BASE + 'rpc.NECPluginV2AgentNotifierApi'
STUB_OFC_DRIVER = ('networking_nec.tests.unit.openflow.'
                   'stub_ofc_driver.StubOFCDriver')


class NecPluginV2TestCaseBase(object):
    _plugin_name = PLUGIN_NAME

    def _set_nec_ini(self):
        nec_config.register_plugin_opts()
        cfg.CONF.set_override('driver', STUB_OFC_DRIVER, group='OFC')
        cfg.CONF.set_override('enable_packet_filter', False, group='OFC')

    def patch_remote_calls(self):
        self.plugin_notifier_p = mock.patch(NOTIFIER)
        self.ofc_manager_p = mock.patch(OFC_MANAGER)
        self.plugin_notifier_p.start()
        self.ofc_manager_p.start()

    def setup_nec_plugin_base(self):
        self._set_nec_ini()
        self.patch_remote_calls()

    def setUp(self, plugin=None, ext_mgr=None):
        plugin = plugin or self._plugin_name
        self._set_nec_ini()
        super(NecPluginV2TestCaseBase, self).setUp(plugin, ext_mgr=ext_mgr)


class NecPluginV2TestCase(NecPluginV2TestCaseBase,
                          base.NeutronDbPluginV2TestCase):

    def rpcapi_update_ports(self, agent_id='nec-q-agent.fake',
                            datapath_id="0xabc", added=[], removed=[]):
        kwargs = {'topic': topics.AGENT,
                  'agent_id': agent_id,
                  'datapath_id': datapath_id,
                  'port_added': added, 'port_removed': removed}
        self.callback_nec.update_ports(self.context, **kwargs)

    def setUp(self, plugin=None, ext_mgr=None):
        super(NecPluginV2TestCase, self).setUp(plugin, ext_mgr=ext_mgr)
        self.plugin = manager.NeutronManager.get_plugin()
        self.plugin.ofc = fake_ofc_manager.patch_ofc_manager()
        self.ofc = self.plugin.ofc
        self.callback_nec = rpc.NECPluginV2RPCCallbacks(self.plugin)
        self.context = context.get_admin_context()
        self.net_create_status = 'ACTIVE'
        self.port_create_status = 'DOWN'


class TestNecBasicGet(test_plugin.TestBasicGet, NecPluginV2TestCase):
    pass


class TestNecV2HTTPResponse(test_plugin.TestV2HTTPResponse,
                            NecPluginV2TestCase):
    pass


class TestNecNetworksV2(test_plugin.TestNetworksV2, NecPluginV2TestCase):
    pass


class TestNecPortsV2Callback(NecPluginV2TestCase):

    def _get_portinfo(self, port_id):
        return ndb.get_portinfo(self.context.session, port_id)

    def test_portinfo_create(self):
        with self.port() as port:
            port_id = port['port']['id']
            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'DOWN')
            self.assertEqual(self.ofc.create_ofc_port.call_count, 0)
            self.assertIsNone(self._get_portinfo(port_id))

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'ACTIVE')
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertIsNotNone(self._get_portinfo(port_id))

            expected = [
                mock.call.exists_ofc_port(mock.ANY, port_id),
                mock.call.create_ofc_port(mock.ANY, port_id, mock.ANY),
            ]
            self.ofc.assert_has_calls(expected)

    def test_portinfo_delete_before_port_deletion(self):
        self._test_portinfo_delete()

    def test_portinfo_delete_after_port_deletion(self):
        self._test_portinfo_delete(portinfo_delete_first=False)

    def _test_portinfo_delete(self, portinfo_delete_first=True):
        with self.port() as port:
            port_id = port['port']['id']
            portinfo = {'id': port_id, 'port_no': 456}
            self.assertEqual(self.ofc.create_ofc_port.call_count, 0)
            self.assertIsNone(self._get_portinfo(port_id))

            self.rpcapi_update_ports(added=[portinfo])
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 0)
            self.assertIsNotNone(self._get_portinfo(port_id))

            # Before port-deletion, switch port removed message is sent.
            if portinfo_delete_first:
                self.rpcapi_update_ports(removed=[port_id])
                self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
                self.assertIsNone(self._get_portinfo(port_id))
        self._delete('ports', port['port']['id'])

        # The port and portinfo is expected to delete when exiting with-clause.
        self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
        self.assertIsNone(self._get_portinfo(port_id))
        if not portinfo_delete_first:
            self.rpcapi_update_ports(removed=[port_id])

        # Ensure port deletion is called once.
        self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
        self.assertIsNone(self._get_portinfo(port_id))

        expected = [
            mock.call.exists_ofc_port(mock.ANY, port_id),
            mock.call.create_ofc_port(mock.ANY, port_id, mock.ANY),
            mock.call.exists_ofc_port(mock.ANY, port_id),
            mock.call.delete_ofc_port(mock.ANY, port_id, mock.ANY),
        ]
        self.ofc.assert_has_calls(expected)

    def test_portinfo_added_unknown_port(self):
        portinfo = {'id': 'dummy-p1', 'port_no': 123}
        self.rpcapi_update_ports(added=[portinfo])
        self.assertIsNone(ndb.get_portinfo(self.context.session,
                                           'dummy-p1'))
        self.assertEqual(self.ofc.exists_ofc_port.call_count, 0)
        self.assertEqual(self.ofc.create_ofc_port.call_count, 0)

    def _test_portinfo_change(self, portinfo_change_first=True):
        with self.port() as port:
            port_id = port['port']['id']
            self.assertEqual(self.ofc.create_ofc_port.call_count, 0)

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 0)
            self.assertEqual(ndb.get_portinfo(self.context.session,
                                              port_id).port_no, 123)

            if portinfo_change_first:
                portinfo = {'id': port_id, 'port_no': 456}
                self.rpcapi_update_ports(added=[portinfo])
                # OFC port is recreated.
                self.assertEqual(self.ofc.create_ofc_port.call_count, 2)
                self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
                self.assertEqual(ndb.get_portinfo(self.context.session,
                                                  port_id).port_no, 456)
        self._delete('ports', port['port']['id'])

        if not portinfo_change_first:
            # The port is expected to delete when exiting with-clause.
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)

            portinfo = {'id': port_id, 'port_no': 456}
            self.rpcapi_update_ports(added=[portinfo])
            # No OFC operations are expected.
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
            self.assertIsNone(ndb.get_portinfo(self.context.session, port_id))

    def test_portinfo_change(self):
        self._test_portinfo_change()

    def test_portinfo_change_for_nonexisting_port(self):
        self._test_portinfo_change(portinfo_change_first=False)

    def test_port_migration(self):
        agent_id_a, datapath_id_a, port_no_a = 'nec-q-agent.aa', '0xaaa', 10
        agent_id_b, datapath_id_b, port_no_b = 'nec-q-agent.bb', '0xbbb', 11

        with self.port() as port:
            port_id = port['port']['id']
            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'DOWN')

            portinfo_a = {'id': port_id, 'port_no': port_no_a}
            self.rpcapi_update_ports(agent_id=agent_id_a,
                                     datapath_id=datapath_id_a,
                                     added=[portinfo_a])

            portinfo_b = {'id': port_id, 'port_no': port_no_b}
            self.rpcapi_update_ports(agent_id=agent_id_b,
                                     datapath_id=datapath_id_b,
                                     added=[portinfo_b])

            self.rpcapi_update_ports(agent_id=agent_id_a,
                                     datapath_id=datapath_id_a,
                                     removed=[port_id])

            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'ACTIVE')
            self.assertTrue(self.ofc.ofc_ports[port_id])

            expected = [
                mock.call.exists_ofc_port(mock.ANY, port_id),
                mock.call.create_ofc_port(mock.ANY, port_id, mock.ANY),
                mock.call.exists_ofc_port(mock.ANY, port_id),
                mock.call.delete_ofc_port(mock.ANY, port_id, mock.ANY),
                mock.call.exists_ofc_port(mock.ANY, port_id),
                mock.call.create_ofc_port(mock.ANY, port_id, mock.ANY),
            ]
            self.ofc.assert_has_calls(expected)
            self.assertEqual(2, self.ofc.create_ofc_port.call_count)
            self.assertEqual(1, self.ofc.delete_ofc_port.call_count)

    def test_portinfo_readd(self):
        with self.port() as port:
            port_id = port['port']['id']
            self.plugin.get_port(self.context, port_id)

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'ACTIVE')
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 0)
            self.assertIsNotNone(self._get_portinfo(port_id))

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            sport = self.plugin.get_port(self.context, port_id)
            self.assertEqual(sport['status'], 'ACTIVE')
            self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_port.call_count, 0)
            self.assertIsNotNone(self._get_portinfo(port_id))


class TestNecPluginOfcManager(NecPluginV2TestCase):
    def setUp(self):
        super(TestNecPluginOfcManager, self).setUp()
        self.ofc = self.plugin.ofc

    def _create_resource(self, resource, data):
        collection = resource + 's'
        data = {resource: data}
        req = self.new_create_request(collection, data)
        res = self.deserialize(self.fmt, req.get_response(self.api))
        return res[resource]

    def _update_resource(self, resource, id, data):
        collection = resource + 's'
        data = {resource: data}
        req = self.new_update_request(collection, data, id)
        res = self.deserialize(self.fmt, req.get_response(self.api))
        return res[resource]

    def _show_resource(self, resource, id):
        collection = resource + 's'
        req = self.new_show_request(collection, id)
        res = self.deserialize(self.fmt, req.get_response(self.api))
        return res[resource]

    def _list_resource(self, resource):
        collection = resource + 's'
        req = self.new_list_request(collection)
        res = req.get_response(self.api)
        return res[collection]

    def _delete_resource(self, resource, id):
        collection = resource + 's'
        req = self.new_delete_request(collection, id)
        res = req.get_response(self.api)
        return res.status_int

    def test_create_network(self):
        net = None
        ctx = mock.ANY
        with self.network() as network:
            net = network['network']
            self.assertEqual(network['network']['status'], 'ACTIVE')
        self._delete('networks', network['network']['id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_create_network_with_admin_state_down(self):
        net = None
        ctx = mock.ANY
        with self.network(admin_state_up=False) as network:
            net = network['network']
            self.assertEqual(network['network']['status'], 'DOWN')
        self._delete('networks', network['network']['id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_create_two_network(self):
        nets = []
        ctx = mock.ANY
        with self.network() as net1:
            nets.append(net1['network'])
            self.assertEqual(net1['network']['status'], 'ACTIVE')
            with self.network() as net2:
                nets.append(net2['network'])
                self.assertEqual(net2['network']['status'], 'ACTIVE')
        self._delete('networks', net2['network']['id'])
        self._delete('networks', net1['network']['id'])
        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, nets[0]['id'],
                                         nets[0]['name']),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, nets[1]['id'],
                                         nets[1]['name']),
            mock.call.exists_ofc_network(ctx, nets[1]['id']),
            mock.call.delete_ofc_network(ctx, nets[1]['id'], mock.ANY),
            mock.call.exists_ofc_network(ctx, nets[0]['id']),
            mock.call.delete_ofc_network(ctx, nets[0]['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_create_network_fail(self):
        self.ofc.create_ofc_network.side_effect = nexc.OFCException(
            reason='hoge')

        net = None
        ctx = mock.ANY
        # NOTE: We don't delete network through api, but db will be cleaned in
        # tearDown(). When OFCManager has failed to create a network on OFC,
        # it does not keeps ofc_network entry and will fail to delete this
        # network from OFC. Deletion of network is not the scope of this test.
        with self.network() as network:
            net = network['network']
            self.assertEqual(net['status'], 'ERROR')
            net_ref = self._show('networks', net['id'])
            self.assertEqual(net_ref['network']['status'], 'ERROR')

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name'])
        ]
        self.ofc.assert_has_calls(expected)

    def test_update_network(self):
        net = None
        ctx = mock.ANY
        with self.network() as network:
            net = network['network']
            self.assertEqual(network['network']['status'], 'ACTIVE')

            net_ref = self._show('networks', net['id'])
            self.assertEqual(net_ref['network']['status'], 'ACTIVE')

            # Set admin_state_up to False
            res = self._update_resource('network', net['id'],
                                        {'admin_state_up': False})
            self.assertFalse(res['admin_state_up'])
            self.assertEqual(res['status'], 'DOWN')

            net_ref = self._show('networks', net['id'])
            self.assertEqual(net_ref['network']['status'], 'DOWN')

            # Set admin_state_up to True
            res = self._update_resource('network', net['id'],
                                        {'admin_state_up': True})
            self.assertTrue(res['admin_state_up'])
            self.assertEqual(res['status'], 'ACTIVE')

            net_ref = self._show('networks', net['id'])
            self.assertEqual(net_ref['network']['status'], 'ACTIVE')
        self._delete('networks', network['network']['id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_create_port_no_ofc_creation(self):
        net = None
        p1 = None
        ctx = mock.ANY
        with self.subnet() as subnet:
            with self.port(subnet=subnet) as port:
                p1 = port['port']
                net_id = port['port']['network_id']
                net = self._show_resource('network', net_id)
                self.assertEqual(net['status'], 'ACTIVE')
                self.assertEqual(p1['status'], 'DOWN')

                p1_ref = self._show('ports', p1['id'])
                self.assertEqual(p1_ref['port']['status'], 'DOWN')
        self._delete('ports', port['port']['id'])
        self._delete('networks', port['port']['network_id'])
        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_create_port_with_ofc_creation(self):
        net = None
        p1 = None
        ctx = mock.ANY
        with self.subnet() as subnet:
            with self.port(subnet=subnet) as port:
                p1 = port['port']
                net_id = port['port']['network_id']
                net = self._show_resource('network', net_id)
                self.assertEqual(net['status'], 'ACTIVE')
                self.assertEqual(p1['status'], 'DOWN')

                p1_ref = self._show('ports', p1['id'])
                self.assertEqual(p1_ref['port']['status'], 'DOWN')

                # Check the port is not created on OFC
                self.assertFalse(self.ofc.create_ofc_port.call_count)

                # Register portinfo, then the port is created on OFC
                portinfo = {'id': p1['id'], 'port_no': 123}
                self.rpcapi_update_ports(added=[portinfo])
                self.assertEqual(self.ofc.create_ofc_port.call_count, 1)

                p1_ref = self._show('ports', p1['id'])
                self.assertEqual(p1_ref['port']['status'], 'ACTIVE')
        self._delete('ports', port['port']['id'])
        self._delete('networks', port['port']['network_id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.create_ofc_port(ctx, p1['id'], mock.ANY),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.delete_ofc_port(ctx, p1['id'], mock.ANY),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_delete_network_with_dhcp_port(self):
        ctx = mock.ANY
        with self.network() as network:
            with self.subnet(network=network):
                net = network['network']
                p = self._create_resource(
                    'port',
                    {'network_id': net['id'],
                     'tenant_id': net['tenant_id'],
                     'device_owner': constants.DEVICE_OWNER_DHCP,
                     'device_id': 'dhcp-port1'})
                # Make sure that the port is created on OFC.
                portinfo = {'id': p['id'], 'port_no': 123}
                self.rpcapi_update_ports(added=[portinfo])
                # In a case of dhcp port, the port is deleted automatically
                # when delete_network.
        self._delete('networks', network['network']['id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id,
                                         net['id'], net['name']),
            mock.call.exists_ofc_port(ctx, p['id']),
            mock.call.create_ofc_port(ctx, p['id'], mock.ANY),
            mock.call.exists_ofc_port(ctx, p['id']),
            mock.call.delete_ofc_port(ctx, p['id'], mock.ANY),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_delete_network_with_error_status(self):
        self.ofc.set_raise_exc('create_ofc_network',
                               nexc.OFCException(reason='fake error'))

        with self.network() as net:
            net_id = net['network']['id']
            net_ref = self._show('networks', net_id)
            self.assertEqual(net_ref['network']['status'], 'ERROR')
        self._delete('networks', net['network']['id'])
        ctx = mock.ANY
        tenant_id = self._tenant_id
        net_name = mock.ANY
        net = mock.ANY
        expected = [
            mock.call.exists_ofc_tenant(ctx, tenant_id),
            mock.call.create_ofc_tenant(ctx, tenant_id),
            mock.call.create_ofc_network(ctx, tenant_id, net_id, net_name),
            mock.call.exists_ofc_network(ctx, net_id),
            mock.call.exists_ofc_tenant(ctx, tenant_id),
            mock.call.delete_ofc_tenant(ctx, tenant_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertFalse(self.ofc.delete_ofc_network.call_count)

    def test_delete_network_with_ofc_deletion_failure(self):
        self.ofc.set_raise_exc('delete_ofc_network',
                               nexc.OFCException(reason='hoge'))

        with self.network() as net:
            net_id = net['network']['id']

            self._delete('networks', net_id,
                         expected_code=webob.exc.HTTPInternalServerError.code)

            net_ref = self._show('networks', net_id)
            self.assertEqual(net_ref['network']['status'], 'ERROR')

            self.ofc.set_raise_exc('delete_ofc_network', None)
        self._delete('networks', net['network']['id'])

        ctx = mock.ANY
        tenant = mock.ANY
        net_name = mock.ANY
        net = mock.ANY
        expected = [
            mock.call.create_ofc_network(ctx, tenant, net_id, net_name),
            mock.call.exists_ofc_network(ctx, net_id),
            mock.call.delete_ofc_network(ctx, net_id, net),
            mock.call.exists_ofc_network(ctx, net_id),
            mock.call.delete_ofc_network(ctx, net_id, net),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_network.call_count, 2)

    def test_delete_network_with_deactivating_auto_delete_port_failure(self):
        self.ofc.set_raise_exc('delete_ofc_port',
                               nexc.OFCException(reason='hoge'))

        with self.network() as net:
            net_id = net['network']['id']

            device_owner = db_base_plugin_v2.AUTO_DELETE_PORT_OWNERS[0]
            port = self._make_port(self.fmt, net_id, device_owner=device_owner)
            port_id = port['port']['id']

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

        self._delete('networks', net_id,
                     expected_code=webob.exc.HTTPInternalServerError.code)

        net_ref = self._show('networks', net_id)
        self.assertEqual(net_ref['network']['status'], 'ACTIVE')
        port_ref = self._show('ports', port_id)
        self.assertEqual(port_ref['port']['status'], 'ERROR')

        self.ofc.set_raise_exc('delete_ofc_port', None)
        self._delete('networks', net_id)

        ctx = mock.ANY
        tenant = mock.ANY
        net_name = mock.ANY
        net = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.create_ofc_network(ctx, tenant, net_id, net_name),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_network(ctx, net_id),
            mock.call.delete_ofc_network(ctx, net_id, net)
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_network.call_count, 1)

    def test_update_port(self):
        self._test_update_port_with_admin_state(resource='port')

    def test_update_network_with_ofc_port(self):
        self._test_update_port_with_admin_state(resource='network')

    def _test_update_port_with_admin_state(self, resource='port'):
        net = None
        p1 = None
        ctx = mock.ANY

        if resource == 'network':
            net_ini_admin_state = False
            port_ini_admin_state = True
        else:
            net_ini_admin_state = True
            port_ini_admin_state = False

        with self.network(admin_state_up=net_ini_admin_state) as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet,
                               admin_state_up=port_ini_admin_state) as port:
                    p1 = port['port']
                    net_id = port['port']['network_id']
                    res_id = net_id if resource == 'network' else p1['id']
                    self.assertEqual(p1['status'], 'DOWN')

                    net = self._show_resource('network', net_id)

                    # Check the port is not created on OFC
                    self.assertFalse(self.ofc.create_ofc_port.call_count)

                    # Register portinfo, then the port is created on OFC
                    portinfo = {'id': p1['id'], 'port_no': 123}
                    self.rpcapi_update_ports(added=[portinfo])
                    self.assertFalse(self.ofc.create_ofc_port.call_count)

                    res = self._update_resource(resource, res_id,
                                                {'admin_state_up': True})
                    self.assertEqual(res['status'], 'ACTIVE')
                    self.assertEqual(self.ofc.create_ofc_port.call_count, 1)
                    self.assertFalse(self.ofc.delete_ofc_port.call_count)

                    res = self._update_resource(resource, res_id,
                                                {'admin_state_up': False})
                    self.assertEqual(res['status'], 'DOWN')
                    self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)
        self._delete('ports', port['port']['id'])
        self._delete('networks', port['port']['network_id'])

        expected = [
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_tenant(ctx, self._tenant_id),
            mock.call.create_ofc_network(ctx, self._tenant_id, net['id'],
                                         net['name']),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.create_ofc_port(ctx, p1['id'], mock.ANY),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.delete_ofc_port(ctx, p1['id'], mock.ANY),

            mock.call.exists_ofc_port(ctx, p1['id']),
            mock.call.exists_ofc_network(ctx, net['id']),
            mock.call.delete_ofc_network(ctx, net['id'], mock.ANY),
            mock.call.exists_ofc_tenant(ctx, self._tenant_id),
            mock.call.delete_ofc_tenant(ctx, self._tenant_id)
        ]
        self.ofc.assert_has_calls(expected)

    def test_update_port_with_ofc_creation_failure(self):
        with self.port(admin_state_up=False) as port:
            port_id = port['port']['id']
            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            self.ofc.set_raise_exc('create_ofc_port',
                                   nexc.OFCException(reason='hoge'))

            body = {'port': {'admin_state_up': True}}
            res = self._update('ports', port_id, body)
            self.assertEqual(res['port']['status'], 'ERROR')
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')

            body = {'port': {'admin_state_up': False}}
            res = self._update('ports', port_id, body)
            self.assertEqual(res['port']['status'], 'ERROR')
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')

            self.ofc.set_raise_exc('create_ofc_port', None)

            body = {'port': {'admin_state_up': True}}
            res = self._update('ports', port_id, body)
            self.assertEqual(res['port']['status'], 'ACTIVE')
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ACTIVE')
        self._delete('ports', port['port']['id'])
        ctx = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.create_ofc_port.call_count, 2)

    def test_update_port_with_ofc_deletion_failure(self):
        with self.port() as port:
            port_id = port['port']['id']
            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            self.ofc.set_raise_exc('delete_ofc_port',
                                   nexc.OFCException(reason='hoge'))

            body = {'port': {'admin_state_up': False}}
            self._update('ports', port_id, body,
                         expected_code=webob.exc.HTTPInternalServerError.code)
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')

            body = {'port': {'admin_state_up': True}}
            res = self._update('ports', port_id, body)
            self.assertEqual(res['port']['status'], 'ERROR')
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')

            self.ofc.set_raise_exc('delete_ofc_port', None)

            body = {'port': {'admin_state_up': False}}
            res = self._update('ports', port_id, body)
            self.assertEqual(res['port']['status'], 'DOWN')
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'DOWN')
        self._delete('ports', port['port']['id'])

        ctx = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_port.call_count, 2)

    def test_delete_port_with_error_status(self):
        self.ofc.set_raise_exc('create_ofc_port',
                               nexc.OFCException(reason='fake'))

        with self.port() as port:
            port_id = port['port']['id']
            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])
            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')
        self._delete('ports', port['port']['id'])

        ctx = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertFalse(self.ofc.delete_ofc_port.call_count)

    def test_delete_port_with_ofc_deletion_failure(self):
        self.ofc.set_raise_exc('delete_ofc_port',
                               nexc.OFCException(reason='hoge'))

        with self.port() as port:
            port_id = port['port']['id']

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            self._delete('ports', port_id,
                         expected_code=webob.exc.HTTPInternalServerError.code)

            port_ref = self._show('ports', port_id)
            self.assertEqual(port_ref['port']['status'], 'ERROR')

            self.ofc.set_raise_exc('delete_ofc_port', None)
        self._delete('ports', port['port']['id'])

        ctx = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port)
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_port.call_count, 2)

    def _test_delete_port_for_disappeared_ofc_port(self, raised_exc):
        self.ofc.set_raise_exc('delete_ofc_port', raised_exc)

        with self.port() as port:
            port_id = port['port']['id']

            portinfo = {'id': port_id, 'port_no': 123}
            self.rpcapi_update_ports(added=[portinfo])

            self._delete('ports', port_id)

            # Check the port on neutron db is deleted. NotFound for
            # neutron port itself should be handled by called. It is
            # consistent with ML2 behavior, but it may need to be
            # revisit.
            self._show('ports', port_id,
                       expected_code=webob.exc.HTTPNotFound.code)

        ctx = mock.ANY
        port = mock.ANY
        expected = [
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.create_ofc_port(ctx, port_id, port),
            mock.call.exists_ofc_port(ctx, port_id),
            mock.call.delete_ofc_port(ctx, port_id, port),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_port.call_count, 1)

    def test_delete_port_for_nonexist_ofc_port(self):
        self._test_delete_port_for_disappeared_ofc_port(
            nexc.OFCResourceNotFound(resource='ofc_port'))

    def test_delete_port_for_noofcmap_ofc_port(self):
        self._test_delete_port_for_disappeared_ofc_port(
            nexc.OFCMappingNotFound(resource='port', neutron_id='port1'))


class TestNecAllowedAddressPairs(NecPluginV2TestCase,
                                 test_pair.TestAllowedAddressPairs):
    pass



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import uuid

import netaddr
from oslo_config import cfg
from oslo_log import helpers as log_helpers

from neutron.api.v2 import attributes
from neutron.common import constants
from neutron.common import exceptions as qexc
from neutron import manager

from networking_nec.plugins.openflow.db import api as ndb
from networking_nec.plugins.openflow.extensions import packetfilter as ext_pf
from networking_nec.plugins.openflow import ofc_client
from networking_nec.plugins.openflow import ofc_driver_base


class InvalidOFCIdFormat(qexc.NeutronException):
    message = _("OFC %(resource)s ID has an invalid format: %(ofc_id)s")


class PFCDriverBase(ofc_driver_base.OFCDriverBase):
    """Base Class for PDC Drivers.

    PFCDriverBase provides methods to handle PFC resources through REST API.
    This uses ofc resource path instead of ofc resource ID.

    The class implements the API for PFC V4.0 or later.
    """

    router_supported = False

    match_ofc_network_id = re.compile(
        "^/tenants/(?P<tenant_id>[^/]+)/networks/(?P<network_id>[^/]+)$")
    match_ofc_port_id = re.compile(
        "^/tenants/(?P<tenant_id>[^/]+)/networks/(?P<network_id>[^/]+)"
        "/ports/(?P<port_id>[^/]+)$")

    def __init__(self, conf_ofc):
        self.client = ofc_client.OFCClient(host=conf_ofc.host,
                                           port=conf_ofc.port,
                                           use_ssl=conf_ofc.use_ssl,
                                           key_file=conf_ofc.key_file,
                                           cert_file=conf_ofc.cert_file,
                                           insecure_ssl=conf_ofc.insecure_ssl)

    @classmethod
    def filter_supported(cls):
        return False

    def _generate_pfc_str(self, raw_str):
        """Generate PFC acceptable String."""
        return re.sub(r'[^0-9a-zA-Z]', '_', raw_str)

    def _generate_pfc_id(self, id_str):
        """Generate ID on PFC.

        Currently, PFC ID must be less than 32.
        Shorten UUID string length from 36 to 31 by follows:
          * delete UUID Version and hyphen (see RFC4122)
          * ensure str length
        """
        try:
            # openstack.common.uuidutils.is_uuid_like() returns
            # False for KeyStone tenant_id, so uuid.UUID is used
            # directly here to accept tenant_id as UUID string
            uuid_str = str(uuid.UUID(id_str)).replace('-', '')
            uuid_no_version = uuid_str[:12] + uuid_str[13:]
            return uuid_no_version[:31]
        except Exception:
            return self._generate_pfc_str(id_str)[:31]

    def _generate_pfc_description(self, desc):
        """Generate Description on PFC.

        Currently, PFC Description must be less than 128.
        """
        return self._generate_pfc_str(desc)[:127]

    def _extract_ofc_network_id(self, ofc_network_id):
        match = self.match_ofc_network_id.match(ofc_network_id)
        if match:
            return match.group('network_id')
        raise InvalidOFCIdFormat(resource='network', ofc_id=ofc_network_id)

    def _extract_ofc_port_id(self, ofc_port_id):
        match = self.match_ofc_port_id.match(ofc_port_id)
        if match:
            return {'tenant': match.group('tenant_id'),
                    'network': match.group('network_id'),
                    'port': match.group('port_id')}
        raise InvalidOFCIdFormat(resource='port', ofc_id=ofc_port_id)

    def create_tenant(self, description, tenant_id=None):
        ofc_tenant_id = self._generate_pfc_id(tenant_id)
        body = {'id': ofc_tenant_id}
        self.client.post('/tenants', body=body)
        return '/tenants/' + ofc_tenant_id

    def delete_tenant(self, ofc_tenant_id):
        return self.client.delete(ofc_tenant_id)

    def create_network(self, ofc_tenant_id, description, network_id=None):
        path = "%s/networks" % ofc_tenant_id
        pfc_desc = self._generate_pfc_description(description)
        body = {'description': pfc_desc}
        res = self.client.post(path, body=body)
        ofc_network_id = res['id']
        return path + '/' + ofc_network_id

    def delete_network(self, ofc_network_id):
        return self.client.delete(ofc_network_id)

    def create_port(self, ofc_network_id, portinfo,
                    port_id=None, filters=None):
        path = "%s/ports" % ofc_network_id
        body = {'datapath_id': portinfo.datapath_id,
                'port': str(portinfo.port_no),
                'vid': str(portinfo.vlan_id)}
        if self.filter_supported() and filters:
            body['filters'] = [self._extract_ofc_filter_id(pf[1])
                               for pf in filters]
        res = self.client.post(path, body=body)
        ofc_port_id = res['id']
        return path + '/' + ofc_port_id

    def delete_port(self, ofc_port_id):
        return self.client.delete(ofc_port_id)


class PFCFilterDriverMixin(object):
    """PFC PacketFilter Driver Mixin."""
    filters_path = "/filters"
    filter_path = "/filters/%s"

    # PFC specific constants
    MIN_PRIORITY = 1
    MAX_PRIORITY = 32766
    CREATE_ONLY_FIELDS = ['action', 'priority']
    PFC_ALLOW_ACTION = "pass"
    PFC_DROP_ACTION = "drop"

    match_ofc_filter_id = re.compile("^/filters/(?P<filter_id>[^/]+)$")

    @classmethod
    def filter_supported(cls):
        return True

    def _set_param(self, filter_dict, body, key, create, convert_to=None):
        if key in filter_dict:
            if filter_dict[key]:
                if convert_to:
                    body[key] = convert_to(filter_dict[key])
                else:
                    body[key] = filter_dict[key]
            elif not create:
                body[key] = ""

    def _extract_ofc_filter_port_id(self, ofc_port_id):
        """Return ofc port id description for packet filter.

        It returns either of the following format:
        {'tenant': xxxx, 'network': xxxx, 'port': xxxx} or
        {'tenant': xxxx, 'router': xxxx, 'interface': xxxx}
        If no matching ofc id is found, InvalidOFCIdFormat is raised.
        """
        if cfg.CONF.OFC.support_packet_filter_on_ofc_router:
            try:
                return self._extract_ofc_router_inf_id(ofc_port_id)
            except InvalidOFCIdFormat:
                pass
        return self._extract_ofc_port_id(ofc_port_id)

    def _generate_body(self, filter_dict, apply_ports=None, create=True):
        body = {}

        if create:
            # action : pass, drop (mandatory)
            if filter_dict['action'].lower() in ext_pf.ALLOW_ACTIONS:
                body['action'] = self.PFC_ALLOW_ACTION
            else:
                body['action'] = self.PFC_DROP_ACTION
            # priority : mandatory
            body['priority'] = filter_dict['priority']

        for key in ['src_mac', 'dst_mac', 'src_port', 'dst_port']:
            self._set_param(filter_dict, body, key, create)

        for key in ['src_cidr', 'dst_cidr']:
            # CIDR must contain netmask even if it is an address.
            convert_to = lambda x: str(netaddr.IPNetwork(x))
            self._set_param(filter_dict, body, key, create, convert_to)

        # protocol : decimal (0-255)
        if 'protocol' in filter_dict:
            if (not filter_dict['protocol'] or
                # In the case of ARP, ip_proto should be set to wildcard.
                # eth_type is set during adding an entry to DB layer.
                    filter_dict['protocol'].lower() == ext_pf.PROTO_NAME_ARP):
                if not create:
                    body['protocol'] = ""
            elif filter_dict['protocol'].lower() == constants.PROTO_NAME_ICMP:
                body['protocol'] = constants.PROTO_NUM_ICMP
            elif filter_dict['protocol'].lower() == constants.PROTO_NAME_TCP:
                body['protocol'] = constants.PROTO_NUM_TCP
            elif filter_dict['protocol'].lower() == constants.PROTO_NAME_UDP:
                body['protocol'] = constants.PROTO_NUM_UDP
            else:
                body['protocol'] = int(filter_dict['protocol'], 0)

        # eth_type : hex (0x0-0xFFFF)
        self._set_param(filter_dict, body, 'eth_type', create, hex)

        # apply_ports
        if apply_ports:
            # each element of apply_ports is a tuple of (neutron_id, ofc_id),
            body['apply_ports'] = []
            for p in apply_ports:
                try:
                    _ofc_id = self._extract_ofc_filter_port_id(p[1])
                    body['apply_ports'].append(_ofc_id)
                except InvalidOFCIdFormat:
                    pass

        return body

    def _validate_filter_common(self, filter_dict):
        # Currently PFC support only IPv4 CIDR.
        for field in ['src_cidr', 'dst_cidr']:
            if (not filter_dict.get(field) or
                    filter_dict[field] == attributes.ATTR_NOT_SPECIFIED):
                continue
            net = netaddr.IPNetwork(filter_dict[field])
            if net.version != 4:
                raise ext_pf.PacketFilterIpVersionNonSupported(
                    version=net.version, field=field, value=filter_dict[field])
        if ('priority' in filter_dict and
            not (self.MIN_PRIORITY <= filter_dict['priority']
                 <= self.MAX_PRIORITY)):
            raise ext_pf.PacketFilterInvalidPriority(
                min=self.MIN_PRIORITY, max=self.MAX_PRIORITY)

    def _validate_duplicate_priority(self, context, filter_dict):
        plugin = manager.NeutronManager.get_plugin()
        filters = {'network_id': [filter_dict['network_id']],
                   'priority': [filter_dict['priority']]}
        ret = plugin.get_packet_filters(context, filters=filters,
                                        fields=['id'])
        if ret:
            raise ext_pf.PacketFilterDuplicatedPriority(
                priority=filter_dict['priority'])

    def validate_filter_create(self, context, filter_dict):
        self._validate_filter_common(filter_dict)
        self._validate_duplicate_priority(context, filter_dict)

    def validate_filter_update(self, context, filter_dict):
        for field in self.CREATE_ONLY_FIELDS:
            if field in filter_dict:
                raise ext_pf.PacketFilterUpdateNotSupported(field=field)
        self._validate_filter_common(filter_dict)

    @log_helpers.log_method_call
    def create_filter(self, context, filter_dict, filter_id=None):
        in_port_id = filter_dict.get('in_port')
        apply_ports = ndb.get_active_ports_on_ofc(
            context, filter_dict['network_id'], in_port_id)

        body = self._generate_body(filter_dict, apply_ports, create=True)
        res = self.client.post(self.filters_path, body=body)
        # filter_id passed from a caller is not used.
        # ofc_filter_id is generated by PFC because the prefix of
        # filter_id has special meaning and it is internally used.
        ofc_filter_id = res['id']
        return self.filter_path % ofc_filter_id

    @log_helpers.log_method_call
    def update_filter(self, ofc_filter_id, filter_dict):
        body = self._generate_body(filter_dict, create=False)
        self.client.put(ofc_filter_id, body)

    @log_helpers.log_method_call
    def delete_filter(self, ofc_filter_id):
        return self.client.delete(ofc_filter_id)

    def _extract_ofc_filter_id(self, ofc_filter_id):
        match = self.match_ofc_filter_id.match(ofc_filter_id)
        if match:
            return match.group('filter_id')
        raise InvalidOFCIdFormat(resource='filter', ofc_id=ofc_filter_id)


class PFCRouterDriverMixin(object):

    router_supported = True
    router_nat_supported = False

    match_ofc_router_inf_id = re.compile(
        "^/tenants/(?P<tenant_id>[^/]+)/routers/(?P<router_id>[^/]+)"
        "/interfaces/(?P<router_inf_id>[^/]+)$")

    def _extract_ofc_router_inf_id(self, ofc_router_inf_id):
        match = self.match_ofc_router_inf_id.match(ofc_router_inf_id)
        if match:
            return {'tenant': match.group('tenant_id'),
                    'router': match.group('router_id'),
                    'interface': match.group('router_inf_id')}
        raise InvalidOFCIdFormat(resource='router-interface',
                                 ofc_id=ofc_router_inf_id)

    def create_router(self, ofc_tenant_id, router_id, description):
        path = '%s/routers' % ofc_tenant_id
        res = self.client.post(path, body=None)
        ofc_router_id = res['id']
        return path + '/' + ofc_router_id

    def delete_router(self, ofc_router_id):
        return self.client.delete(ofc_router_id)

    def add_router_interface(self, ofc_router_id, ofc_net_id,
                             ip_address=None, mac_address=None):
        # ip_address : <ip_address>/<netmask> (e.g., 10.0.0.0/24)
        path = '%s/interfaces' % ofc_router_id
        body = {'net_id': self._extract_ofc_network_id(ofc_net_id)}
        if ip_address:
            body['ip_address'] = ip_address
        if mac_address:
            body['mac_address'] = mac_address
        res = self.client.post(path, body=body)
        return path + '/' + res['id']

    def update_router_interface(self, ofc_router_inf_id,
                                ip_address=None, mac_address=None):
        # ip_address : <ip_address>/<netmask> (e.g., 10.0.0.0/24)
        if not ip_address and not mac_address:
            return
        body = {}
        if ip_address:
            body['ip_address'] = ip_address
        if mac_address:
            body['mac_address'] = mac_address
        return self.client.put(ofc_router_inf_id, body=body)

    def delete_router_interface(self, ofc_router_inf_id):
        return self.client.delete(ofc_router_inf_id)

    def list_router_routes(self, ofc_router_id):
        path = '%s/routes' % ofc_router_id
        ret = self.client.get(path)
        # Prepend ofc_router_id to route_id
        for r in ret['routes']:
            r['id'] = ofc_router_id + '/routes/' + r['id']
        return ret['routes']

    def add_router_route(self, ofc_router_id, destination, nexthop):
        path = '%s/routes' % ofc_router_id
        body = {'destination': destination,
                'nexthop': nexthop}
        ret = self.client.post(path, body=body)
        return path + '/' + ret['id']

    def delete_router_route(self, ofc_router_route_id):
        return self.client.delete(ofc_router_route_id)


class PFCV3Driver(PFCDriverBase):

    def create_tenant(self, description, tenant_id):
        ofc_tenant_id = self._generate_pfc_id(tenant_id)
        return "/tenants/" + ofc_tenant_id

    def delete_tenant(self, ofc_tenant_id):
        pass


class PFCV4Driver(PFCDriverBase):
    pass


class PFCV5Driver(PFCRouterDriverMixin, PFCDriverBase):
    pass


class PFCV51Driver(PFCFilterDriverMixin, PFCV5Driver):
    pass



# Copyright 2013, NEC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.api.v2 import attributes
from neutron import context
from neutron.extensions import securitygroup as ext_sg
from neutron import manager
from neutron.tests.unit.agent import test_securitygroups_rpc as test_sg_rpc
from neutron.tests.unit.extensions import test_securitygroup as test_sg

from networking_nec.tests.unit.openflow import test_plugin


PLUGIN_NAME = test_plugin.PLUGIN_NAME
OFC_MANAGER = 'networking_nec.plugins.openflow.ofc_manager.OFCManager'
NOTIFIER = 'networking_nec.plugins.openflow.rpc.NECPluginV2AgentNotifierApi'


class NecSecurityGroupsTestCase(test_sg.SecurityGroupDBTestCase):

    def setUp(self, plugin=None):
        test_sg_rpc.set_firewall_driver(test_sg_rpc.FIREWALL_HYBRID_DRIVER)
        mock.patch(NOTIFIER).start()
        mock.patch(OFC_MANAGER).start()
        self._attribute_map_bk_ = {}
        for item in attributes.RESOURCE_ATTRIBUTE_MAP:
            self._attribute_map_bk_[item] = (attributes.
                                             RESOURCE_ATTRIBUTE_MAP[item].
                                             copy())
        super(NecSecurityGroupsTestCase, self).setUp(PLUGIN_NAME)
        plugin = manager.NeutronManager.get_plugin()
        self.notifier = plugin.notifier
        self.rpc = plugin.callback_sg

    def tearDown(self):
        super(NecSecurityGroupsTestCase, self).tearDown()
        attributes.RESOURCE_ATTRIBUTE_MAP = self._attribute_map_bk_


class TestNecSGServerRpcCallBack(
        test_sg_rpc.SGServerRpcCallBackTestCase,
        NecSecurityGroupsTestCase):
    pass


class TestNecSecurityGroups(NecSecurityGroupsTestCase,
                            test_sg.TestSecurityGroups,
                            test_sg_rpc.SGNotificationTestMixin):

    def setUp(self):
        super(TestNecSecurityGroups, self).setUp()
        self.ctx = context.get_admin_context()

    def test_security_group_get_port_from_device(self):
        with self.network() as n, \
                self.security_group() as sg:
            with self.subnet(n):
                res = self._create_port(self.fmt, n['network']['id'])
                port = self.deserialize(self.fmt, res)
                port_id = port['port']['id']
                sg_id = sg['security_group']['id']
                fixed_ips = port['port']['fixed_ips']

                data = {'port': {'fixed_ips': fixed_ips,
                                 'name': port['port']['name'],
                                 ext_sg.SECURITYGROUPS: [sg_id]}}
                req = self.new_update_request('ports', data, port_id)
                res = self.deserialize(self.fmt,
                                       req.get_response(self.api))

                plugin = manager.NeutronManager.get_plugin()
                port_dict = plugin.get_port_from_device(self.ctx, port_id)
                self.assertEqual(port_id, port_dict['id'])
                self.assertEqual([sg_id],
                                 port_dict[ext_sg.SECURITYGROUPS])
                self.assertEqual([], port_dict['security_group_rules'])
                self.assertEqual([fixed_ips[0]['ip_address']],
                                 port_dict['fixed_ips'])
                self._delete('ports', port_id)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
import requests

from networking_nec._i18n import _, _LI, _LW
from networking_nec.plugins.openflow import exceptions as nexc


LOG = logging.getLogger(__name__)


class OFCClient(object):
    """A HTTP/HTTPS client for OFC Drivers."""

    def __init__(self, host="127.0.0.1", port=8888, use_ssl=False,
                 key_file=None, cert_file=None, insecure_ssl=False):
        """Creates a new client to some OFC.

        :param host: The host where service resides
        :param port: The port where service resides
        :param use_ssl: True to use SSL, False to use HTTP
        :param key_file: The SSL key file to use if use_ssl is true
        :param cert_file: The SSL cert file to use if use_ssl is true
        :param insecure_ssl: Don't verify SSL certificate
        """
        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self.key_file = key_file
        self.cert_file = cert_file
        self.insecure_ssl = insecure_ssl
        self.connection = None

    def _format_error_message(self, status, detail):
        detail = ' ' + detail if detail else ''
        return (_("Operation on OFC failed: %(status)s%(msg)s") %
                {'status': status, 'msg': detail})

    def _get_response(self, method, action, body=None):
            headers = {"Content-Type": "application/json"}
            protocol = "http"
            certs = {'key_file': self.key_file, 'cert_file': self.cert_file}
            certs = dict((x, certs[x]) for x in certs if certs[x] is not None)
            verify = True

            if self.use_ssl:
                protocol = "https"
                if self.insecure_ssl:
                    verify = False

            url = "%s://%s:%d%s" % (protocol, self.host, int(self.port),
                                    action)

            res = requests.request(method, url, data=body, headers=headers,
                                   cert=certs, verify=verify)
            return res

    def do_single_request(self, method, action, body=None):
        action = cfg.CONF.OFC.path_prefix + action
        LOG.debug("Client request: %(host)s:%(port)s "
                  "%(method)s %(action)s [%(body)s]",
                  {'host': self.host, 'port': self.port,
                   'method': method, 'action': action, 'body': body})
        if isinstance(body, dict):
            body = jsonutils.dumps(body)
        try:
            res = self._get_response(method, action, body)
            data = res.text
            LOG.debug("OFC returns [%(status)s:%(data)s]",
                      {'status': res.status_code,
                       'data': data})

            # Try to decode JSON data if possible.
            try:
                data = jsonutils.loads(data)
            except (ValueError, TypeError):
                pass

            if res.status_code in (requests.codes.OK,
                                   requests.codes.CREATED,
                                   requests.codes.ACCEPTED,
                                   requests.codes.NO_CONTENT):
                return data
            elif res.status_code == requests.codes.SERVICE_UNAVAILABLE:
                retry_after = res.headers.get('retry-after')
                LOG.warning(_LW("OFC returns ServiceUnavailable "
                                "(retry-after=%s)"), retry_after)
                raise nexc.OFCServiceUnavailable(retry_after=retry_after)
            elif res.status_code == requests.codes.NOT_FOUND:
                LOG.info(_LI("Specified resource %s does not exist on OFC "),
                         action)
                raise nexc.OFCResourceNotFound(resource=action)
            else:
                LOG.warning(_LW("Operation on OFC failed: "
                                "status=%(status)s, detail=%(detail)s"),
                            {'status': res.status_code, 'detail': data})
                params = {'reason': _("Operation on OFC failed"),
                          'status': res.status_code}
                if isinstance(data, dict):
                    params['err_code'] = data.get('err_code')
                    params['err_msg'] = data.get('err_msg')
                else:
                    params['err_msg'] = data
                raise nexc.OFCException(**params)
        except requests.exceptions.RequestException as e:
            reason = _("Failed to connect OFC : %s") % e
            LOG.error(reason)
            raise nexc.OFCException(reason=reason)

    def do_request(self, method, action, body=None):
        max_attempts = cfg.CONF.OFC.api_max_attempts
        for i in range(max_attempts, 0, -1):
            try:
                return self.do_single_request(method, action, body)
            except nexc.OFCServiceUnavailable as e:
                with excutils.save_and_reraise_exception() as ctxt:
                    try:
                        wait_time = int(e.retry_after)
                    except (ValueError, TypeError):
                        wait_time = None
                    if i > 1 and wait_time:
                        LOG.info(_LI("Waiting for %s seconds due to "
                                     "OFC Service_Unavailable."), wait_time)
                        time.sleep(wait_time)
                        ctxt.reraise = False
                        continue

    def get(self, action):
        return self.do_request("GET", action)

    def post(self, action, body=None):
        return self.do_request("POST", action, body=body)

    def put(self, action, body=None):
        return self.do_request("PUT", action, body=body)

    def delete(self, action):
        return self.do_request("DELETE", action)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base
from oslo_config import cfg

from neutron.agent.common import config as agent_config

from networking_nec.plugins.openflow import config


class ConfigurationTest(base.BaseTestCase):

    def setUp(self):
        super(ConfigurationTest, self).setUp()
        # NOTE: I am not sure why register_agent_opts is already called
        # with a different *Opt object, but it causes register_*_opts to fail.
        # Resetting cfg.CONF to a fresh ConfigOpts helps us.
        # This test class tests the default value loaded by register_*_opts,
        # so I believe it is good to reset loaded options before the tests.
        self.patch(cfg, 'CONF', cfg.ConfigOpts())

    def test_defaults_plugin_opts(self):
        config.register_plugin_opts()

        self.assertEqual('127.0.0.1', cfg.CONF.OFC.host)
        self.assertEqual('8888', cfg.CONF.OFC.port)
        # Check path_prefix is an empty string explicitly.
        self.assertEqual('', cfg.CONF.OFC.path_prefix)
        self.assertEqual('trema', cfg.CONF.OFC.driver)
        self.assertTrue(cfg.CONF.OFC.enable_packet_filter)
        self.assertTrue(cfg.CONF.OFC.support_packet_filter_on_ofc_router)
        self.assertFalse(cfg.CONF.OFC.use_ssl)
        self.assertIsNone(cfg.CONF.OFC.key_file)
        self.assertIsNone(cfg.CONF.OFC.cert_file)

    def test_defaults_agent_opts(self):
        config.register_agent_opts()
        agent_config.register_root_helper(cfg.CONF)

        self.assertEqual('br-int', cfg.CONF.OVS.integration_bridge)
        self.assertEqual(2, cfg.CONF.AGENT.polling_interval)
        self.assertEqual('sudo', cfg.CONF.AGENT.root_helper)



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import mock
from oslo_config import cfg
import webob.exc

from neutron.api.v2 import attributes
from neutron import context
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_db_plugin

from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow.extensions import packetfilter as ext_pf
from networking_nec.tests.unit.openflow import test_plugin


class PacketfilterExtensionManager(ext_pf.Packetfilter):

    @classmethod
    def get_resources(cls):
        # Add the resources to the global attribute map
        # This is done here as the setup process won't
        # initialize the main API router which extends
        # the global attribute map
        attributes.RESOURCE_ATTRIBUTE_MAP.update(
            {'packet_filters': ext_pf.PACKET_FILTER_ATTR_MAP})
        return super(PacketfilterExtensionManager, cls).get_resources()


class TestNecPluginPacketFilterBase(test_plugin.NecPluginV2TestCase):

    def _set_nec_ini(self):
        super(TestNecPluginPacketFilterBase, self)._set_nec_ini()
        cfg.CONF.set_override('enable_packet_filter', True, group='OFC')

    def setUp(self):
        ext_mgr = PacketfilterExtensionManager()
        super(TestNecPluginPacketFilterBase, self).setUp(ext_mgr=ext_mgr)

    def _create_packet_filter(self, fmt, net_id, expected_res_status=None,
                              arg_list=None, **kwargs):
        data = {'packet_filter': {'network_id': net_id,
                                  'tenant_id': self._tenant_id,
                                  'priority': '1',
                                  'action': 'ALLOW'}}

        for arg in (('name', 'admin_state_up', 'action', 'priority', 'in_port',
                     'src_mac', 'dst_mac', 'eth_type', 'src_cidr', 'dst_cidr',
                     'protocol', 'src_port', 'dst_port') +
                    (arg_list or ())):
            # Arg must be present
            if arg in kwargs:
                data['packet_filter'][arg] = kwargs[arg]
        pf_req = self.new_create_request('packet_filters', data, fmt)
        if (kwargs.get('set_context') and 'tenant_id' in kwargs):
            # create a specific auth context for this request
            pf_req.environ['neutron.context'] = context.Context(
                '', kwargs['tenant_id'])

        pf_res = pf_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(pf_res.status_int, expected_res_status)
        return pf_res

    def _make_packet_filter(self, fmt, net_id, expected_res_status=None,
                            **kwargs):
        res = self._create_packet_filter(fmt, net_id, expected_res_status,
                                         **kwargs)
        # Things can go wrong - raise HTTP exc with res code only
        # so it can be caught by unit tests
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(fmt, res)

    @contextlib.contextmanager
    def packet_filter_on_network(self, network=None, fmt=None, **kwargs):
        with test_db_plugin.optional_ctx(network,
                                         self.network) as network_to_use:
            net_id = network_to_use['network']['id']
            pf = self._make_packet_filter(fmt or self.fmt, net_id, **kwargs)
            yield pf
            if not network:
                self._delete('networks', network_to_use['network']['id'])

    @contextlib.contextmanager
    def packet_filter_on_port(self, port=None, fmt=None, set_portinfo=True,
                              **kwargs):
        with test_db_plugin.optional_ctx(port, self.port) as port_to_use:
            net_id = port_to_use['port']['network_id']
            port_id = port_to_use['port']['id']

            if set_portinfo:
                portinfo = {'id': port_id,
                            'port_no': kwargs.get('port_no', 123)}
                kw = {'added': [portinfo]}
                if 'datapath_id' in kwargs:
                    kw['datapath_id'] = kwargs['datapath_id']
                self.rpcapi_update_ports(**kw)

            kwargs['in_port'] = port_id
            pf = self._make_packet_filter(fmt or self.fmt, net_id, **kwargs)
            self.assertEqual(port_id, pf['packet_filter']['in_port'])
            yield pf


class TestNecPluginPacketFilter(TestNecPluginPacketFilterBase):

    def setUp(self):
        super(TestNecPluginPacketFilter, self).setUp()
        # Remove attributes explicitly from mock object to check
        # a case where there are no update_filter and validate_*.
        del self.ofc.driver.update_filter
        del self.ofc.driver.validate_filter_create
        del self.ofc.driver.validate_filter_update

    def test_list_packet_filters(self):
        self._list('packet_filters')

    def test_create_pf_on_network_no_ofc_creation(self):
        with self.packet_filter_on_network(admin_state_up=False) as pf:
            self.assertEqual(pf['packet_filter']['status'], 'DOWN')

        self.assertFalse(self.ofc.create_ofc_packet_filter.called)
        self.assertFalse(self.ofc.delete_ofc_packet_filter.called)

    def test_create_pf_on_port_no_ofc_creation(self):
        with self.packet_filter_on_port(admin_state_up=False,
                                        set_portinfo=False) as pf:
            self.assertEqual(pf['packet_filter']['status'], 'DOWN')

        self.assertFalse(self.ofc.create_ofc_packet_filter.called)
        self.assertFalse(self.ofc.delete_ofc_packet_filter.called)

    def test_create_pf_on_network_with_ofc_creation(self):
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']
            self.assertEqual(pf['packet_filter']['status'], 'ACTIVE')

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 1)

    def test_create_pf_on_port_with_ofc_creation(self):
        with self.packet_filter_on_port() as pf:
            pf_id = pf['packet_filter']['id']
            self.assertEqual(pf['packet_filter']['status'], 'ACTIVE')
            self._delete('packet_filters', pf_id)

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 1)

    def _test_create_pf_with_protocol(self, protocol, expected_eth_type):
        with self.packet_filter_on_network(protocol=protocol) as pf:
            pf_data = pf['packet_filter']
            self.assertEqual(protocol, pf_data['protocol'])
            self.assertEqual(expected_eth_type, pf_data['eth_type'])

    def test_create_pf_with_protocol_tcp(self):
        self._test_create_pf_with_protocol('TCP', 0x800)

    def test_create_pf_with_protocol_udp(self):
        self._test_create_pf_with_protocol('UDP', 0x800)

    def test_create_pf_with_protocol_icmp(self):
        self._test_create_pf_with_protocol('ICMP', 0x800)

    def test_create_pf_with_protocol_arp(self):
        self._test_create_pf_with_protocol('ARP', 0x806)

    def test_create_pf_with_inconsistent_protocol_and_eth_type(self):
        with self.packet_filter_on_network(protocol='TCP') as pf:
            pf_data = pf['packet_filter']
            pf_id = pf_data['id']
            self.assertEqual('TCP', pf_data['protocol'])
            self.assertEqual(0x800, pf_data['eth_type'])
            data = {'packet_filter': {'eth_type': 0x806}}
            self._update('packet_filters', pf_id, data,
                         expected_code=409)

    def test_create_pf_with_invalid_priority(self):
        with self.network() as net:
            net_id = net['network']['id']
            kwargs = {'priority': 'high'}
            self._create_packet_filter(self.fmt, net_id,
                                       webob.exc.HTTPBadRequest.code,
                                       **kwargs)
        self.assertFalse(self.ofc.create_ofc_packet_filter.called)

    def test_create_pf_with_ofc_creation_failure(self):
        self.ofc.set_raise_exc('create_ofc_packet_filter',
                               nexc.OFCException(reason='hoge'))

        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ERROR')

            self.ofc.set_raise_exc('create_ofc_packet_filter', None)

            # Retry activate packet_filter (even if there is no change).
            data = {'packet_filter': {}}
            self._update('packet_filters', pf_id, data)

            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ACTIVE')

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 2)

    def test_show_pf_on_network(self):
        kwargs = {
            'name': 'test-pf-net',
            'admin_state_up': False,
            'action': 'DENY',
            'priority': '102',
            'src_mac': '00:11:22:33:44:55',
            'dst_mac': '66:77:88:99:aa:bb',
            'eth_type': '2048',
            'src_cidr': '192.168.1.0/24',
            'dst_cidr': '192.168.2.0/24',
            'protocol': 'TCP',
            'src_port': '35001',
            'dst_port': '22'
        }

        with self.packet_filter_on_network(**kwargs) as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)

            # convert string to int.
            kwargs.update({'priority': 102, 'eth_type': 2048,
                           'src_port': 35001, 'dst_port': 22,
                           'in_port': None})

            self.assertEqual(pf_id, pf_ref['packet_filter']['id'])
            for key in kwargs:
                self.assertEqual(kwargs[key], pf_ref['packet_filter'][key])

    def test_show_pf_on_network_with_wildcards(self):
        kwargs = {
            'name': 'test-pf-net',
            'admin_state_up': False,
            'action': 'DENY',
            'priority': '102',
        }

        with self.packet_filter_on_network(**kwargs) as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)

            # convert string to int.
            kwargs.update({'priority': 102,
                           'in_port': None,
                           'src_mac': None,
                           'dst_mac': None,
                           'eth_type': None,
                           'src_cidr': None,
                           'dst_cidr': None,
                           'protocol': None,
                           'src_port': None,
                           'dst_port': None})

            self.assertEqual(pf_id, pf_ref['packet_filter']['id'])
            for key in kwargs:
                self.assertEqual(kwargs[key], pf_ref['packet_filter'][key])

    def test_show_pf_on_port(self):
        kwargs = {
            'name': 'test-pf-port',
            'admin_state_up': False,
            'action': 'DENY',
            'priority': '0o147',
            'src_mac': '00:11:22:33:44:55',
            'dst_mac': '66:77:88:99:aa:bb',
            'eth_type': 2048,
            'src_cidr': '192.168.1.0/24',
            'dst_cidr': '192.168.2.0/24',
            'protocol': 'TCP',
            'dst_port': '0x50'
        }

        with self.packet_filter_on_port(**kwargs) as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)

            # convert string to int.
            kwargs.update({'priority': 103, 'eth_type': 2048,
                           'dst_port': 80,
                           # wildcard field is None in a response.
                           'src_port': None})

            self.assertEqual(pf_id, pf_ref['packet_filter']['id'])
            self.assertTrue(pf_ref['packet_filter']['in_port'])
            for key in kwargs:
                self.assertEqual(kwargs[key], pf_ref['packet_filter'][key])

    def test_show_pf_not_found(self):
        pf_id = '00000000-ffff-ffff-ffff-000000000000'

        self._show('packet_filters', pf_id,
                   expected_code=webob.exc.HTTPNotFound.code)

    def test_update_pf_on_network(self):
        ctx = mock.ANY
        pf_dict = mock.ANY
        with self.packet_filter_on_network(admin_state_up=False) as pf:
            pf_id = pf['packet_filter']['id']

            self.assertFalse(self.ofc.create_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': True}}
            self._update('packet_filters', pf_id, data)
            self.ofc.create_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id, pf_dict)

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)

    def test_update_pf_on_port(self):
        ctx = mock.ANY
        pf_dict = mock.ANY
        with self.packet_filter_on_port(admin_state_up=False) as pf:
            pf_id = pf['packet_filter']['id']

            self.assertFalse(self.ofc.create_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': True}}
            self._update('packet_filters', pf_id, data)
            self.ofc.create_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id, pf_dict)

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)

    def test_delete_pf_with_error_status(self):
        self.ofc.set_raise_exc('create_ofc_packet_filter',
                               nexc.OFCException(reason='fake'))
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ERROR')

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(1, self.ofc.create_ofc_packet_filter.call_count)
        self.assertEqual(0, self.ofc.delete_ofc_packet_filter.call_count)

    def test_activate_pf_on_port_triggered_by_update_port(self):
        ctx = mock.ANY
        pf_dict = mock.ANY
        self.ofc.set_raise_exc('create_ofc_packet_filter',
                               nexc.PortInfoNotFound(id='fake_id'))
        with self.packet_filter_on_port(set_portinfo=False) as pf:
            pf_id = pf['packet_filter']['id']
            in_port_id = pf['packet_filter']['in_port']

            # create_ofc_packet_filter is now called even when
            # in_port does not exists yet. In this case
            # PortInfoNotFound exception is raised.
            self.assertEqual(1, self.ofc.create_ofc_packet_filter.call_count)
            portinfo = {'id': in_port_id, 'port_no': 123}
            kw = {'added': [portinfo]}
            self.ofc.set_raise_exc('create_ofc_packet_filter', None)
            self.rpcapi_update_ports(**kw)
            self.assertEqual(2, self.ofc.create_ofc_packet_filter.call_count)
            self.ofc.assert_has_calls([
                mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            ])

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            kw = {'removed': [in_port_id]}
            self.rpcapi_update_ports(**kw)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)

        # Ensure pf was created before in_port has activated.
        ctx = mock.ANY
        pf_dict = mock.ANY
        port_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            mock.call.exists_ofc_port(ctx, in_port_id),
            mock.call.create_ofc_port(ctx, in_port_id, port_dict),

            mock.call.exists_ofc_port(ctx, in_port_id),
            mock.call.delete_ofc_port(ctx, in_port_id, port_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(2, self.ofc.create_ofc_packet_filter.call_count)
        self.assertEqual(1, self.ofc.delete_ofc_packet_filter.call_count)

    def test_activate_pf_while_exists_on_ofc(self):
        ctx = mock.ANY
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']

            self.ofc.set_raise_exc('delete_ofc_packet_filter',
                                   nexc.OFCException(reason='hoge'))

            # This update request will make plugin reactivate pf.
            data = {'packet_filter': {'priority': 1000}}
            self._update('packet_filters', pf_id, data,
                         expected_code=webob.exc.HTTPInternalServerError.code)

            self.ofc.set_raise_exc('delete_ofc_packet_filter', None)

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 2)

    def test_deactivate_pf_with_ofc_deletion_failure(self):
        ctx = mock.ANY
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']

            self.ofc.set_raise_exc('delete_ofc_packet_filter',
                                   nexc.OFCException(reason='hoge'))

            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data,
                         expected_code=webob.exc.HTTPInternalServerError.code)

            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ERROR')

            self.ofc.set_raise_exc('delete_ofc_packet_filter', None)

            data = {'packet_filter': {'priority': 1000}}
            self._update('packet_filters', pf_id, data)

            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'DOWN')

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 2)

    def test_delete_pf_with_ofc_deletion_failure(self):
        self.ofc.set_raise_exc('delete_ofc_packet_filter',
                               nexc.OFCException(reason='hoge'))

        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']

            self._delete('packet_filters', pf_id,
                         expected_code=webob.exc.HTTPInternalServerError.code)

            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ERROR')

            self.ofc.set_raise_exc('delete_ofc_packet_filter', None)
            # Then, self._delete('packet_filters', pf_id) will success.

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),

            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 2)

    def test_auto_delete_pf_in_network_deletion(self):
        with self.packet_filter_on_network(admin_state_up=False) as pf:
            pf_id = pf['packet_filter']['id']

        self._show('packet_filters', pf_id,
                   expected_code=webob.exc.HTTPNotFound.code)

    def test_auto_delete_pf_in_port_deletion(self):
        with self.port() as port:
            network = self._show('networks', port['port']['network_id'])

            with self.packet_filter_on_network(network=network) as pfn:
                with self.packet_filter_on_port(port=port,
                                                set_portinfo=False) as pf:
                    pf_id = pf['packet_filter']['id']
                    in_port_id = pf['packet_filter']['in_port']

                    self._delete('ports', in_port_id)
                    # Check the packet filter on the port is deleted.
                    self._show('packet_filters', pf_id,
                               expected_code=webob.exc.HTTPNotFound.code)
                    # Check the packet filter on the network is not deleted.
                    self._show('packet_filters', pfn['packet_filter']['id'])

    def test_no_pf_activation_while_port_operations(self):
        with self.packet_filter_on_port() as pf:
            in_port_id = pf['packet_filter']['in_port']
            self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 0)

            data = {'port': {'admin_state_up': False}}
            self._update('ports', in_port_id, data)
            self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 0)

            data = {'port': {'admin_state_up': True}}
            self._update('ports', in_port_id, data)
            self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
            self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 0)


class TestNecPluginPacketFilterWithValidate(TestNecPluginPacketFilterBase):

    def setUp(self):
        super(TestNecPluginPacketFilterWithValidate, self).setUp()
        # Remove attributes explicitly from mock object to check
        # a case where there are no update_filter.
        del self.ofc.driver.update_filter
        self.validate_create = self.ofc.driver.validate_filter_create
        self.validate_update = self.ofc.driver.validate_filter_update

    def test_create_pf_on_network(self):
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']
            self.assertEqual(pf['packet_filter']['status'], 'ACTIVE')

        ctx = mock.ANY
        pf_dict = mock.ANY
        expected = [
            mock.call.driver.validate_filter_create(ctx, pf_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.create_ofc_packet_filter(ctx, pf_id, pf_dict),
            mock.call.exists_ofc_packet_filter(ctx, pf_id),
            mock.call.delete_ofc_packet_filter(ctx, pf_id),
        ]
        self.ofc.assert_has_calls(expected)
        self.assertEqual(self.ofc.create_ofc_packet_filter.call_count, 1)
        self.assertEqual(self.ofc.delete_ofc_packet_filter.call_count, 1)

    def test_update_pf_on_network(self):
        ctx = mock.ANY
        pf_dict = mock.ANY
        with self.packet_filter_on_network(admin_state_up=False) as pf:
            pf_id = pf['packet_filter']['id']

            self.assertFalse(self.ofc.create_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': True}}
            self._update('packet_filters', pf_id, data)
            self.ofc.create_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id, pf_dict)
            self.ofc.driver.validate_filter_update.assert_called_once_with(
                ctx, data['packet_filter'])

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)
            self.assertEqual(
                2, self.ofc.driver.validate_filter_update.call_count)

    def test_create_pf_on_network_with_validation_error(self):
        self.validate_create.side_effect = ext_pf.PacketFilterInvalidPriority(
            min=1, max=65535)
        with self.network() as net:
            net_id = net['network']['id']
            e = self.assertRaises(webob.exc.HTTPClientError,
                                  self._make_packet_filter,
                                  self.fmt, net_id, expected_res_status=400)
            self.assertEqual(400, e.status_int)

    def test_update_pf_on_network_with_validation_error(self):
        self.validate_update.side_effect = (
            ext_pf.PacketFilterUpdateNotSupported(field='priority'))
        with self.packet_filter_on_network() as pf:
            pf_id = pf['packet_filter']['id']
            pf_ref = self._show('packet_filters', pf_id)
            self.assertEqual(pf_ref['packet_filter']['status'], 'ACTIVE')
            data = {'packet_filter': {'priority': 1000}}
            self._update('packet_filters', pf_id, data,
                         expected_code=400)


class TestNecPluginPacketFilterWithFilterUpdate(TestNecPluginPacketFilterBase):

    def setUp(self):
        super(TestNecPluginPacketFilterWithFilterUpdate, self).setUp()
        # Remove attributes explicitly from mock object to check
        # a case where there are no update_filter and validate_*.
        del self.ofc.driver.validate_filter_create
        del self.ofc.driver.validate_filter_update

    def test_update_pf_toggle_admin_state(self):
        ctx = mock.ANY
        pf_dict = mock.ANY
        with self.packet_filter_on_network(admin_state_up=False) as pf:
            pf_id = pf['packet_filter']['id']

            self.assertFalse(self.ofc.create_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': True}}
            self._update('packet_filters', pf_id, data)
            self.ofc.create_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id, pf_dict)

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)

    def test_update_pf_change_field(self):
        ctx = mock.ANY
        with self.packet_filter_on_network(admin_state_up=True) as pf:
            pf_id = pf['packet_filter']['id']
            self.assertTrue(self.ofc.create_ofc_packet_filter.called)

            data = {'packet_filter': {'src_mac': '12:34:56:78:9a:bc'}}
            self._update('packet_filters', pf_id, data)
            self.ofc.update_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id, data['packet_filter'])
            self.assertEqual(1, self.ofc.update_ofc_packet_filter.call_count)

            self.assertFalse(self.ofc.delete_ofc_packet_filter.called)
            data = {'packet_filter': {'admin_state_up': False}}
            self._update('packet_filters', pf_id, data)
            self.ofc.delete_ofc_packet_filter.assert_called_once_with(
                ctx, pf_id)

            data = {'packet_filter': {'src_mac': '11:22:33:44:55:66'}}
            self._update('packet_filters', pf_id, data)
            self.assertEqual(1, self.ofc.update_ofc_packet_filter.call_count)

            data = {'packet_filter': {'admin_state_up': True}}
            self._update('packet_filters', pf_id, data)

            data = {'packet_filter': {'src_mac': '66:55:44:33:22:11'}}
            self._update('packet_filters', pf_id, data)
            self.assertEqual(2, self.ofc.update_ofc_packet_filter.call_count)



# Copyright 2012-2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from networking_nec._i18n import _LE
from networking_nec.plugins.openflow.db import packetfilter as pf_db
from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow import utils


LOG = logging.getLogger(__name__)


class PacketFilterMixin(pf_db.PacketFilterDbMixin):
    """Mixin class to add packet filter to NECPluginV2."""

    @property
    def packet_filter_enabled(self):
        if not hasattr(self, '_packet_filter_enabled'):
            self._packet_filter_enabled = (
                cfg.CONF.OFC.enable_packet_filter and
                self.ofc.driver.filter_supported())
        return self._packet_filter_enabled

    def remove_packet_filter_extension_if_disabled(self, aliases):
        if not self.packet_filter_enabled:
            LOG.debug('Disabled packet-filter extension.')
            aliases.remove('packet-filter')

    def create_packet_filter(self, context, packet_filter):
        """Create a new packet_filter entry on DB, then try to activate it."""
        LOG.debug("create_packet_filter() called, packet_filter=%s .",
                  packet_filter)

        if hasattr(self.ofc.driver, 'validate_filter_create'):
            pf = packet_filter['packet_filter']
            self.ofc.driver.validate_filter_create(context, pf)
        pf = super(PacketFilterMixin, self).create_packet_filter(
            context, packet_filter)

        return self.activate_packet_filter_if_ready(context, pf)

    def update_packet_filter(self, context, id, packet_filter):
        """Update packet_filter entry on DB, and recreate it if changed.

        If any rule of the packet_filter was changed, recreate it on OFC.
        """
        LOG.debug("update_packet_filter() called, "
                  "id=%(id)s packet_filter=%(packet_filter)s .",
                  {'id': id, 'packet_filter': packet_filter})

        pf_data = packet_filter['packet_filter']
        if hasattr(self.ofc.driver, 'validate_filter_update'):
            self.ofc.driver.validate_filter_update(context, pf_data)

        # validate ownership
        pf_old = self.get_packet_filter(context, id)

        pf = super(PacketFilterMixin, self).update_packet_filter(
            context, id, packet_filter)

        def _packet_filter_changed(old_pf, new_pf):
            LOG.debug('old_pf=%(old_pf)s, new_pf=%(new_pf)s',
                      {'old_pf': old_pf, 'new_pf': new_pf})
            # When the status is ERROR, force sync to OFC.
            if old_pf['status'] == pf_db.PF_STATUS_ERROR:
                LOG.debug('update_packet_filter: Force filter update '
                          'because the previous status is ERROR.')
                return True
            for key in new_pf:
                if key in ('id', 'name', 'tenant_id', 'network_id',
                           'in_port', 'status'):
                    continue
                if old_pf[key] != new_pf[key]:
                    return True
            return False

        if _packet_filter_changed(pf_old, pf):
            if hasattr(self.ofc.driver, 'update_filter'):
                # admin_state is changed
                if pf_old['admin_state_up'] != pf['admin_state_up']:
                    LOG.debug('update_packet_filter: admin_state '
                              'is changed to %s', pf['admin_state_up'])
                    if pf['admin_state_up']:
                        self.activate_packet_filter_if_ready(context, pf)
                    else:
                        self.deactivate_packet_filter(context, pf)
                elif pf['admin_state_up']:
                    LOG.debug('update_packet_filter: admin_state is '
                              'unchanged (True)')
                    if self.ofc.exists_ofc_packet_filter(context, id):
                        pf = self._update_packet_filter(context, pf, pf_data)
                    else:
                        pf = self.activate_packet_filter_if_ready(context, pf)
                else:
                    LOG.debug('update_packet_filter: admin_state is unchanged '
                              '(False). No need to update OFC filter.')
            else:
                pf = self.deactivate_packet_filter(context, pf)
                pf = self.activate_packet_filter_if_ready(context, pf)

        return pf

    def _update_packet_filter(self, context, new_pf, pf_data):
        pf_id = new_pf['id']
        prev_status = new_pf['status']
        try:
            # If previous status is ERROR, try to sync all attributes.
            pf = new_pf if prev_status == pf_db.PF_STATUS_ERROR else pf_data
            self.ofc.update_ofc_packet_filter(context, pf_id, pf)
            new_status = pf_db.PF_STATUS_ACTIVE
            if new_status != prev_status:
                utils.update_resource_status(context, "packet_filter",
                                             pf_id, new_status)
                new_pf['status'] = new_status
            return new_pf
        except Exception as exc:
            with excutils.save_and_reraise_exception():
                if (isinstance(exc, nexc.OFCException) or
                        isinstance(exc, nexc.OFCConsistencyBroken)):
                    LOG.error(_LE("Failed to create packet_filter id=%(id)s "
                                  "on OFC: %(exc)s"),
                              {'id': pf_id, 'exc': exc})
                new_status = pf_db.PF_STATUS_ERROR
                if new_status != prev_status:
                    utils.update_resource_status(context, "packet_filter",
                                                 pf_id, new_status)

    def delete_packet_filter(self, context, id):
        """Deactivate and delete packet_filter."""
        LOG.debug("delete_packet_filter() called, id=%s .", id)

        # validate ownership
        pf = self.get_packet_filter(context, id)

        # deactivate_packet_filter() raises an exception
        # if an error occurs during processing.
        pf = self.deactivate_packet_filter(context, pf)

        super(PacketFilterMixin, self).delete_packet_filter(context, id)

    def activate_packet_filter_if_ready(self, context, packet_filter):
        """Activate packet_filter by creating filter on OFC if ready.

        Conditions to create packet_filter on OFC are:
            * packet_filter admin_state is UP
            * (if 'in_port' is specified) portinfo is available
        """
        LOG.debug("activate_packet_filter_if_ready() called, "
                  "packet_filter=%s.", packet_filter)

        pf_id = packet_filter['id']
        current = packet_filter['status']

        pf_status = current
        if not packet_filter['admin_state_up']:
            LOG.debug("activate_packet_filter_if_ready(): skip pf_id=%s, "
                      "packet_filter.admin_state_up is False.", pf_id)
        elif self.ofc.exists_ofc_packet_filter(context, packet_filter['id']):
            LOG.debug("_activate_packet_filter_if_ready(): skip, "
                      "ofc_packet_filter already exists.")
        else:
            LOG.debug("activate_packet_filter_if_ready(): create "
                      "packet_filter id=%s on OFC.", pf_id)
            try:
                self.ofc.create_ofc_packet_filter(context, pf_id,
                                                  packet_filter)
                pf_status = pf_db.PF_STATUS_ACTIVE
            except nexc.PortInfoNotFound:
                LOG.debug("Skipped to create a packet filter pf_id=%s "
                          "on OFC, no portinfo for the in_port.", pf_id)
            except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
                LOG.error(_LE("Failed to create packet_filter id=%(id)s on "
                              "OFC: %(exc)s"), {'id': pf_id, 'exc': exc})
                pf_status = pf_db.PF_STATUS_ERROR

        if pf_status != current:
            utils.update_resource_status(context, "packet_filter", pf_id,
                                         pf_status)
            packet_filter.update({'status': pf_status})

        return packet_filter

    def deactivate_packet_filter(self, context, packet_filter):
        """Deactivate packet_filter by deleting filter from OFC if exixts."""
        LOG.debug("deactivate_packet_filter_if_ready() called, "
                  "packet_filter=%s.", packet_filter)
        pf_id = packet_filter['id']

        if not self.ofc.exists_ofc_packet_filter(context, pf_id):
            LOG.debug("deactivate_packet_filter(): skip, "
                      "Not found OFC Mapping for packet_filter id=%s.",
                      pf_id)
            return packet_filter

        LOG.debug("deactivate_packet_filter(): "
                  "deleting packet_filter id=%s from OFC.", pf_id)
        try:
            self.ofc.delete_ofc_packet_filter(context, pf_id)
            utils.update_resource_status_if_changed(
                context, "packet_filter", packet_filter, pf_db.PF_STATUS_DOWN)
            return packet_filter
        except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Failed to delete packet_filter id=%(id)s "
                              "from OFC: %(exc)s"),
                          {'id': pf_id, 'exc': exc})
                utils.update_resource_status_if_changed(
                    context, "packet_filter", packet_filter,
                    pf_db.PF_STATUS_ERROR)

    def activate_packet_filters_by_port(self, context, port_id):
        if not self.packet_filter_enabled:
            return

        filters = {'in_port': [port_id], 'admin_state_up': [True],
                   'status': [pf_db.PF_STATUS_DOWN]}
        pfs = self.get_packet_filters(context, filters=filters)
        for pf in pfs:
            self.activate_packet_filter_if_ready(context, pf)

    def deactivate_packet_filters_by_port(self, context, port_id,
                                          raise_exc=True):
        if not self.packet_filter_enabled:
            return

        filters = {'in_port': [port_id], 'status': [pf_db.PF_STATUS_ACTIVE]}
        pfs = self.get_packet_filters(context, filters=filters)
        error = False
        for pf in pfs:
            try:
                self.deactivate_packet_filter(context, pf)
            except (nexc.OFCException, nexc.OFCMappingNotFound):
                error = True
        if raise_exc and error:
            raise nexc.OFCException(
                reason=_('Error occurred while disabling packet '
                         'filter(s) for port %s') % port_id)

    def get_packet_filters_for_port(self, context, port):
        if self.packet_filter_enabled:
            return super(PacketFilterMixin,
                         self).get_packet_filters_for_port(context, port)
